AccessTypeDict = {
    "@graph": [
        {
            "@id": "dandi:AccessType",
            "@type": "rdfs:Class",
            "rdfs:comment": "An enumeration of access status options",
            "rdfs:label": "Access status type",
            "rdfs:subClassOf": {"@id": "schema:Enumeration"},
        },
        {
            "@id": "dandi:Open",
            "@type": "dandi:AccessType",
            "rdfs:comment": "The dandiset is openly accessible",
            "rdfs:label": "Open access",
        },
        {
            "@id": "dandi:Embargoed",
            "@type": "dandi:AccessType",
            "rdfs:comment": "The dandiset is embargoed",
            "rdfs:label": "Embargoed",
        },
        {
            "@id": "dandi:Restricted",
            "@type": "dandi:AccessType",
            "rdfs:comment": "The dandiset is restricted",
            "rdfs:label": "Restricted",
        },
    ]
}

DigestTypeDict = {
    "@graph": [
        {
            "@id": "dandi:DigestType",
            "@type": "rdfs:Class",
            "rdfs:comment": "An enumeration of checksum types",
            "rdfs:label": "Checksum Type",
            "rdfs:subClassOf": {"@id": "schema:Enumeration"},
        },
        {
            "@id": "dandi:md5",
            "@type": "dandi:DigestType",
            "rdfs:comment": "MD5 checksum",
            "rdfs:label": "MD5",
        },
        {
            "@id": "dandi:sha1",
            "@type": "dandi:DigestType",
            "rdfs:comment": "SHA1 checksum",
            "rdfs:label": "SHA1",
        },
        {
            "@id": "dandi:sha2-256",
            "@type": "dandi:DigestType",
            "rdfs:comment": "SHA2-256 checksum",
            "rdfs:label": "SHA2-256",
        },
        {
            "@id": "dandi:sha3-256",
            "@type": "dandi:DigestType",
            "rdfs:comment": "SHA3-256 checksum",
            "rdfs:label": "SHA3-256",
        },
        {
            "@id": "dandi:blake2b-256",
            "@type": "dandi:DigestType",
            "rdfs:comment": "BLAKE2B-256 checksum",
            "rdfs:label": "BLAKE2B-256",
        },
        {
            "@id": "dandi:blake3",
            "@type": "dandi:DigestType",
            "rdfs:comment": "BLAKE3-256 checksum",
            "rdfs:label": "BLAKE3-256",
        },
        {
            "@id": "dandi:dandi-etag",
            "@type": "dandi:DigestType",
            "rdfs:comment": "S3-style ETag",
            "rdfs:label": "DANDI ETag",
        },
    ]
}

IdentifierTypeDict = {
    "@graph": [
        {
            "@id": "dandi:IdentifierType",
            "@type": "rdfs:Class",
            "rdfs:comment": "An enumeration of identifiers",
            "rdfs:label": "License type",
            "rdfs:subClassOf": {"@id": "schema:Enumeration"},
        },
        {
            "@id": "dandi:doi",
            "sameAs": "idorg:doi",
            "@type": "dandi:IdentifierType",
            "rdfs:label": "DOI",
        },
        {
            "@id": "dandi:orcid",
            "sameAs": "idorg:orcid",
            "@type": "dandi:IdentifierType",
            "rdfs:label": "ORCID",
        },
        {
            "@id": "dandi:ror",
            "sameAs": "https://ror.org/",
            "@type": "dandi:IdentifierType",
            "rdfs:label": "ROR",
        },
        {
            "@id": "dandi:dandi",
            "sameAs": "idorg:dandi",
            "@type": "dandi:IdentifierType",
            "rdfs:label": "DANDI",
        },
        {
            "@id": "dandi:rrid",
            "sameAs": "idorg:rrid",
            "@type": "dandi:IdentifierType",
            "rdfs:label": "RRID",
        },
    ]
}
LicenseTypeDict = {
    "@graph": [
        {
            "@id": "dandi:LicenseType",
            "@type": "rdfs:Class",
            "rdfs:comment": "An enumeration of supported licenses",
            "rdfs:label": "License type",
            "rdfs:subClassOf": {"@id": "schema:Enumeration"},
        },
        {
            "@id": "spdx:CC0-1.0",
            "rdfs:seeAlso": "https://creativecommons.org/publicdomain/zero/1.0/legalcode",
            "@type": "dandi:LicenseType",
            "rdfs:label": "CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
        },
        {
            "@id": "spdx:CC-BY-4.0",
            "rdfs:seeAlso": "https://creativecommons.org/licenses/by/4.0/legalcode",
            "@type": "dandi:LicenseType",
            "rdfs:label": "Attribution 4.0 International (CC BY 4.0)",
        },
        {
            "@id": "spdx:CC-BY-NC-4.0",
            "rdfs:seeAlso": "https://creativecommons.org/licenses/by-nc/4.0/legalcode",
            "@type": "dandi:LicenseType",
            "rdfs:label": "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)",
        },
    ]
}

RelationTypeDict = {
    "@graph": [
        {
            "@id": "dandi:RelationType",
            "@type": "rdfs:Class",
            "rdfs:comment": "An enumeration of resource relations",
            "rdfs:label": "Resource relation type",
            "rdfs:subClassOf": {"@id": "schema:Enumeration"},
            "prov:wasDerivedFrom": (
                "https://schema.datacite.org/meta/"
                "kernel-4.2/doc/DataCite-MetadataKernel_v4.2.pdf"
            ),
        },
        {
            "@id": "dandi:IsCitedBy",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates that B includes A in a citation",
            "rdfs:label": "IsCitedBy",
        },
        {
            "@id": "dandi:Cites",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates that A includes B in a citation",
            "rdfs:label": "Cites",
        },
        {
            "@id": "dandi:IsSupplementTo",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates that A is a supplement to B",
            "rdfs:label": "IsSupplementTo",
        },
        {
            "@id": "dandi:IsSupplementedBy",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates that B is a supplement to A",
            "rdfs:label": "IsSupplementedBy",
        },
        {
            "@id": "dandi:IsContinuedBy",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is continued by the work B",
            "rdfs:label": "IsContinuedBy",
        },
        {
            "@id": "dandi:Continues",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is a continuation of the work B",
            "rdfs:label": "Continues",
        },
        {
            "@id": "dandi:Describes",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A describes B",
            "rdfs:label": "Describes",
        },
        {
            "@id": "dandi:IsDescribedBy",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is described by B",
            "rdfs:label": "IsDescribedBy",
        },
        {
            "@id": "dandi:HasMetadata",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates resource A has additional metadata B",
            "rdfs:label": "HasMetadata",
        },
        {
            "@id": "dandi:IsMetadataFor",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates additional metadata A for a resource B",
            "rdfs:label": "IsMetadataFor",
        },
        {
            "@id": "dandi:HasVersion",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A has a version (B)",
            "rdfs:label": "HasVersion",
        },
        {
            "@id": "dandi:IsVersionOf",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is a version of B",
            "rdfs:label": "IsVersionOf",
        },
        {
            "@id": "dandi:IsNewVersionOf",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is a new edition of B",
            "rdfs:label": "IsNewVersionOf",
        },
        {
            "@id": "dandi:IsPreviousVersionOf",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is a previous edition of B",
            "rdfs:label": "IsPreviousVersionOf",
        },
        {
            "@id": "dandi:IsPartOf",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is a portion of B",
            "rdfs:label": "IsPartOf",
        },
        {
            "@id": "dandi:HasPart",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A includes the part B",
            "rdfs:label": "HasPart",
        },
        {
            "@id": "dandi:IsReferencedBy",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is used as a source of information by B",
            "rdfs:label": "IsReferencedBy",
        },
        {
            "@id": "dandi:References",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates B is used as a source of information for A",
            "rdfs:label": "References",
        },
        {
            "@id": "dandi:IsDocumentedBy",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates B is documentation about/explaining A",
            "rdfs:label": "IsDocumentedBy",
        },
        {
            "@id": "dandi:Documents",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is documentation about B",
            "rdfs:label": "Documents",
        },
        {
            "@id": "dandi:IsCompiledBy",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates B is used to compile or create A",
            "rdfs:label": "IsCompiledBy",
        },
        {
            "@id": "dandi:Compiles",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates B is the result of a compile or creation event using A",
            "rdfs:label": "Compiles",
        },
        {
            "@id": "dandi:IsVariantFormOf",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is a variant or different form of B",
            "rdfs:label": "IsVariantFormOf",
        },
        {
            "@id": "dandi:IsOriginalFormOf",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is the original form of B",
            "rdfs:label": "IsOriginalFormOf",
        },
        {
            "@id": "dandi:IsIdenticalTo",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates that A is identical to B",
            "rdfs:label": "IsIdenticalTo",
        },
        {
            "@id": "dandi:IsReviewedBy",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates that A is reviewed by B",
            "rdfs:label": "IsReviewedBy",
        },
        {
            "@id": "dandi:Reviews",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates that A is a review of B",
            "rdfs:label": "Reviews",
        },
        {
            "@id": "dandi:IsDerivedFrom",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates B is a source upon which A is based",
            "rdfs:label": "IsDerivedFrom",
        },
        {
            "@id": "dandi:IsSourceOf",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is a source upon which B is based",
            "rdfs:label": "IsSourceOf",
        },
        {
            "@id": "dandi:IsRequiredBy",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is required by B",
            "rdfs:label": "IsRequiredBy",
        },
        {
            "@id": "dandi:Requires",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A requires B",
            "rdfs:label": "Requires",
        },
        {
            "@id": "dandi:Obsoletes",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A replaces B",
            "rdfs:label": "Obsoletes",
        },
        {
            "@id": "dandi:IsObsoletedBy",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates A is replaced by B",
            "rdfs:label": "IsObsoletedBy",
        },
    ]
}

ParticipantRelationTypeDict = {
    "@graph": [
        {
            "@id": "dandi:ParticipantRelationType",
            "@type": "rdfs:Class",
            "rdfs:comment": "An enumeration of participant relations",
            "rdfs:label": "Participant relation type",
            "rdfs:subClassOf": {"@id": "schema:Enumeration"},
            "prov:wasDerivedFrom": (
                "https://www.ebi.ac.uk/biosamples/docs/guides/relationships"
            ),
        },
        {
            "@id": "dandi:IsChildOf",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates that A is a child of B",
            "rdfs:label": "Child of",
        },
        {
            "@id": "dandi:isParentOf",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates that A is a parent of B",
            "rdfs:label": "Parent of",
        },
        {
            "@id": "dandi:IsSiblingOf",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates that A is a sibling of B",
            "rdfs:label": "Sibling of",
        },
        {
            "@id": "dandi:IsMonozygoticTwinOf",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates that A is a monozygotic twin of B",
            "rdfs:label": "Monozygotic twin of",
        },
        {
            "@id": "dandi:IsDizygoticTwinOf",
            "@type": "dandi:RelationType",
            "rdfs:comment": "Indicates that A is a dizygotic twin of B",
            "rdfs:label": "Dizygotic twin of",
        },
    ]
}

RoleTypeDict = {
    "@graph": [
        {
            "@id": "dandi:RoleType",
            "@type": "rdfs:Class",
            "rdfs:comment": "An enumeration of roles",
            "rdfs:label": "Role Type",
            "rdfs:subClassOf": {"@id": "schema:Enumeration"},
        },
        {
            "@id": "dandi:Author",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Author",
            "rdfs:label": "Author",
        },
        {
            "@id": "dandi:Conceptualization",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Conceptualization",
            "rdfs:label": "Conceptualization",
        },
        {
            "@id": "dandi:ContactPerson",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Contact Person",
            "rdfs:label": "Contact Person",
        },
        {
            "@id": "dandi:DataCollector",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Data Collector",
            "rdfs:label": "Data Collector",
        },
        {
            "@id": "dandi:DataCurator",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Data Curator",
            "rdfs:label": "Data Curator",
        },
        {
            "@id": "dandi:DataManager",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Data Manager",
            "rdfs:label": "Data Manager",
        },
        {
            "@id": "dandi:FormalAnalysis",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Formal Analysis",
            "rdfs:label": "Formal Analysis",
        },
        {
            "@id": "dandi:FundingAcquisition",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Funding Acquisition",
            "rdfs:label": "Funding Acquisition",
        },
        {
            "@id": "dandi:Investigation",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Investigation",
            "rdfs:label": "Investigation",
        },
        {
            "@id": "dandi:Maintainer",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Maintainer",
            "rdfs:label": "Maintainer",
        },
        {
            "@id": "dandi:Methodology",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Methodology",
            "rdfs:label": "Methodology",
        },
        {
            "@id": "dandi:Producer",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Producer",
            "rdfs:label": "Producer",
        },
        {
            "@id": "dandi:ProjectLeader",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Project Leader",
            "rdfs:label": "Project Leader",
        },
        {
            "@id": "dandi:ProjectManager",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Project Manager",
            "rdfs:label": "Project Manager",
        },
        {
            "@id": "dandi:ProjectMember",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Project Member",
            "rdfs:label": "Project Member",
        },
        {
            "@id": "dandi:ProjectAdministration",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Project Administration",
            "rdfs:label": "Project Administration",
        },
        {
            "@id": "dandi:Researcher",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Researcher",
            "rdfs:label": "Researcher",
        },
        {
            "@id": "dandi:Resources",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Resources",
            "rdfs:label": "Resources",
        },
        {
            "@id": "dandi:Software",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Software",
            "rdfs:label": "Software",
        },
        {
            "@id": "dandi:Supervision",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Supervision",
            "rdfs:label": "Supervision",
        },
        {
            "@id": "dandi:Validation",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Validation",
            "rdfs:label": "Validation",
        },
        {
            "@id": "dandi:Visualization",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Visualization",
            "rdfs:label": "Visualization",
        },
        {
            "@id": "dandi:Funder",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Funder",
            "rdfs:label": "Funder",
        },
        {
            "@id": "dandi:Sponsor",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Sponsor",
            "rdfs:label": "Sponsor",
        },
        {
            "@id": "dandi:StudyParticipant",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Participant in a study",
            "rdfs:label": "Study participant",
        },
        {
            "@id": "dandi:Affiliation",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Affiliated with an entity",
            "rdfs:label": "Affiliation",
        },
        {
            "@id": "dandi:EthicsApproval",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Approved ethics protocol",
            "rdfs:label": "Ethics approval",
        },
        {
            "@id": "dandi:Other",
            "@type": "dandi:RoleType",
            "rdfs:comment": "Other",
            "rdfs:label": "Other",
        },
    ]
}



from collections import defaultdict
import os
import os.path as op

import click

from .base import lgr, map_to_click_exceptions
from ..dandiarchive import DandisetURL, _dandi_url_parser, parse_dandi_url
from ..utils import is_url

# TODO: all the recursion options etc


# The use of f-strings apparently makes this not a proper docstring, and so
# click doesn't use it unless we explicitly assign it to `help`:
@click.command(
    help=f"""\
List .nwb files and dandisets metadata.

\b
{_dandi_url_parser.known_patterns}
"""
)
@click.option(
    "-F",
    "--fields",
    help="Comma-separated list of fields to display. "
    "An empty value to trigger a list of "
    "available fields to be printed out",
)
@click.option(
    "-f",
    "--format",
    help="Choose the format/frontend for output. If 'auto', 'pyout' will be "
    "used in case of multiple files, and 'yaml' for a single file.",
    type=click.Choice(["auto", "pyout", "json", "json_pp", "yaml"]),
    default="auto",
)
@click.option(
    "-r",
    "--recursive",
    help="Recurse into content of dandisets/directories. Only .nwb files will "
    "be considered.",
    is_flag=True,
)
@click.option(
    "-J",
    "--jobs",
    help="Number of parallel download jobs.",
    default=6,  # TODO: come up with smart auto-scaling etc
    show_default=True,
)
@click.option(
    "--metadata",
    type=click.Choice(["api", "all", "assets"]),
    default="api",
)
@click.option(
    "--schema",
    help="Convert metadata to new schema version",
    metavar="VERSION",
)
@click.argument("paths", nargs=-1, type=click.Path(exists=False, dir_okay=True))
@map_to_click_exceptions
def ls(paths, schema, metadata, fields=None, format="auto", recursive=False, jobs=6):
    """ List .nwb files and dandisets metadata. """

    # TODO: more logical ordering in case of fields = None
    from .formatter import JSONFormatter, PYOUTFormatter, YAMLFormatter
    from ..consts import metadata_all_fields

    # TODO: avoid
    from ..support.pyout import PYOUT_SHORT_NAMES_rev
    from ..utils import find_files

    common_fields = ("path", "size")
    if schema is not None:
        from ..models import AssetMeta, DandisetMeta

        all_fields = tuple(
            sorted(
                set(common_fields)
                | DandisetMeta.__fields__.keys()
                | AssetMeta.__fields__.keys()
            )
        )
    else:
        all_fields = tuple(sorted(set(common_fields + metadata_all_fields)))

    if fields is not None:
        if fields.strip() == "":
            display_known_fields(all_fields)
            return

        fields = fields.split(",")
        # Map possibly present short names back to full names
        fields = [PYOUT_SHORT_NAMES_rev.get(f.lower(), f) for f in fields]
        unknown_fields = set(fields).difference(all_fields)
        if unknown_fields:
            display_known_fields(all_fields)
            raise click.UsageError(
                "Following fields are not known: %s" % ", ".join(unknown_fields)
            )

    urls = map(is_url, paths)
    # Actually I do not see why and it could be useful to compare local-vs-remote
    # if any(urls) and not all(urls):
    #     raise ValueError(f"ATM cannot mix URLs with local paths. Got {paths}")

    def assets_gen():
        for path in paths:
            if is_url(path):
                parsed_url = parse_dandi_url(path)
                with parsed_url.navigate(
                    include_metadata=metadata in ("all", "assets")
                ) as (client, dandiset, assets):
                    if isinstance(parsed_url, DandisetURL):
                        rec = {
                            "path": dandiset.pop("dandiset", {}).get(
                                "identifier", "ERR#%s" % id(dandiset)
                            )
                        }
                        # flatten the metadata into record to display
                        # rec.update(dandiset.get('metadata', {}))
                        rec.update(dandiset)
                        yield rec
                    if not isinstance(parsed_url, DandisetURL) or recursive:
                        yield from assets
            else:
                # For now we support only individual files
                yield path
                if recursive:
                    yield from find_files(r"\.nwb\Z", path)

    if format == "auto":
        format = "yaml" if any(urls) or (len(paths) == 1 and not recursive) else "pyout"

    if format == "pyout":
        if fields and fields[0] != "path":
            # we must always have path - our "id"
            fields = ["path"] + fields
        out = PYOUTFormatter(fields=fields, wait_for_top=3, max_workers=jobs)
    elif format == "json":
        out = JSONFormatter()
    elif format == "json_pp":
        out = JSONFormatter(indent=2)
    elif format == "yaml":
        out = YAMLFormatter()
    else:
        raise NotImplementedError("Unknown format %s" % format)

    async_keys = set(all_fields)
    if fields is not None:
        async_keys = async_keys.intersection(fields)
    async_keys = tuple(async_keys.difference(common_fields))

    errors = defaultdict(list)  # problem: [] paths
    with out:
        for asset in assets_gen():
            if isinstance(asset, str):  # path
                rec = {}
                rec["path"] = asset

                try:
                    if (not fields or "size" in fields) and not op.isdir(asset):
                        rec["size"] = os.stat(asset).st_size

                    if async_keys:
                        cb = get_metadata_ls(
                            asset,
                            async_keys,
                            errors=errors,
                            flatten=format == "pyout",
                            schema=schema,
                        )
                        if format == "pyout":
                            rec[async_keys] = cb
                        else:
                            # TODO: parallel execution
                            # For now just call callback and get all the fields
                            cb_res = cb()
                            # TODO: we should stop masking exceptions in get_metadata_ls,
                            # and centralize logic regardless either it is for pyout or not
                            # and do parallelizaion on our end, so at large it is
                            if cb_res is None:
                                raise
                            for k, v in cb_res.items():
                                rec[k] = v
                except Exception as exc:
                    _add_exc_error(asset, rec, errors, exc)
            elif isinstance(asset, dict):
                # ready record
                if schema is not None and asset.get("schemaVersion") != schema:
                    raise NotImplementedError(
                        "Record conversion between schema versions is not"
                        " implemented.  Found schemaVersion="
                        f"{asset.get('schemaVersion')} where {schema} was"
                        " requested"
                    )
                # TODO: harmonization for pyout
                rec = asset
            else:
                raise TypeError(asset)

            if not rec:
                errors["Empty record"].append(asset)
                lgr.debug("Skipping a record for %s since empty", asset)
                continue
            out(rec)
    if errors:
        lgr.warning(
            "Failed to operate on some paths (empty records were listed):\n %s",
            "\n ".join("%s: %d paths" % (k, len(v)) for k, v in errors.items()),
        )


def _add_exc_error(asset, rec, errors, exc):
    """A helper to centralize collection of errors for pyout and non-pyout reporting"""
    lgr.debug("Problem obtaining metadata for %s: %s", asset, exc)
    errors[str(type(exc).__name__)].append(asset)
    rec["errors"] = rec.get("errors", 0) + 1


def display_known_fields(all_fields):
    from ..support.pyout import PYOUT_SHORT_NAMES

    # Display all known fields
    click.secho("Known fields:")
    for field in all_fields:
        s = "- " + field
        if field in PYOUT_SHORT_NAMES:
            s += " or %s" % PYOUT_SHORT_NAMES[field]
        click.secho(s)
    return


def flatten_v(v):
    """Return while flattening nested lists/dicts

    lists and tuples would get items converted to strings and joined
    with ", " separator.

    dicts would get items represented as "key: value" before flattening
    a list of them.
    """
    if isinstance(v, (tuple, list)):
        return ", ".join(map(str, map(flatten_v, v)))
    elif isinstance(v, dict):
        return flatten_v(["%s: %s" % i for i in v.items()])
    return v


def flatten_meta_to_pyout_v1(meta):
    """Given a meta record, possibly flatten record since no nested records
    supported yet

    lists become joined using ', ', dicts get individual key: values.
    lists of dict - doing nothing magical.

    Empty values are not considered.

    Parameters
    ----------
    meta: dict
    """
    out = {}

    # normalize some fields and remove completely empty
    for f, v in (meta or dict()).items():
        if not v:
            continue
        if isinstance(v, dict):
            for vf, vv in flatten_meta_to_pyout_v1(v).items():
                out["%s_%s" % (f, vf)] = flatten_v(vv)
        else:
            out[f] = flatten_v(v)
    return out


def flatten_meta_to_pyout(meta):
    """Given a meta record, possibly flatten record since no nested records
    supported yet

    lists become joined using ', ', dicts become lists of "key: value" strings first.
    lists of dict - doing nothing magical.

    Empty values are not considered.

    Parameters
    ----------
    meta: dict
    """
    out = {}

    # normalize some fields and remove completely empty
    for f, v in (meta or dict()).items():
        if not v:
            continue
        out[f] = flatten_v(v)
    return out


def get_metadata_ls(path, keys, errors, flatten=False, schema=None):
    from ..dandiset import APIDandiset
    from ..metadata import get_metadata, nwb2asset
    from ..pynwb_utils import get_nwb_version, ignore_benign_pynwb_warnings

    ignore_benign_pynwb_warnings()

    def fn():
        rec = {}
        # No need for calling get_metadata if no keys are needed from it
        if keys is None or list(keys) != ["nwb_version"]:
            try:
                if schema is not None:
                    if op.isdir(path):
                        dandiset = APIDandiset(path, schema_version=schema)
                        rec = dandiset.metadata
                    else:
                        rec = nwb2asset(path, schema_version=schema).json_dict()
                else:
                    rec = get_metadata(path)
            except Exception as exc:
                _add_exc_error(path, rec, errors, exc)
            if flatten:
                rec = flatten_meta_to_pyout(rec)
        if keys is not None:
            rec = {k: v for k, v in rec.items() if k in keys}
        if (
            not op.isdir(path)
            and "nwb_version" not in rec
            and (keys and "nwb_version" in keys)
        ):
            # Let's at least get that one
            try:
                rec["nwb_version"] = get_nwb_version(path)
            except Exception as exc:
                _add_exc_error(path, rec, errors, exc)
        return rec

    return fn



import os
import os.path as op

import click

from .base import dandiset_path_option, devel_debug_option, lgr, map_to_click_exceptions
from ..consts import file_operation_modes


@click.command()
@dandiset_path_option(
    help="A top directory (local) of the dandiset to organize files under. "
    "If not specified, dandiset current directory is under is assumed. "
    "For 'simulate' mode target dandiset/directory must not exist.",
    type=click.Path(dir_okay=True, file_okay=False),
)
@click.option(
    "--invalid",
    help="What to do if files without sufficient metadata are encountered.",
    type=click.Choice(["fail", "warn"]),
    default="fail",
    show_default=True,
)
@click.option(
    "-f",
    "--files-mode",
    help="If 'dry' - no action is performed, suggested renames are printed. "
    "If 'simulate' - hierarchy of empty files at --local-top-path is created. "
    "Note that previous layout should be removed prior this operation.  "
    "If 'auto' - whichever of symlink, hardlink, copy is allowed by system. "
    "The other modes (copy, move, symlink, hardlink) define how data files "
    "should be made available.",
    type=click.Choice(file_operation_modes),
    default="auto",
    show_default=True,
)
@click.argument("paths", nargs=-1, type=click.Path(exists=True))
@devel_debug_option()
@map_to_click_exceptions
def organize(
    paths,
    dandiset_path=None,
    invalid="fail",
    files_mode="auto",
    devel_debug=False,
):
    """(Re)organize files according to the metadata.

    The purpose of this command is to take advantage of metadata contained in
    the .nwb files to provide datasets with consistently named files, so their
    naming reflects data they contain.

    .nwb files are organized into a hierarchy of subfolders one per each
    "subject", e.g. sub-0001 if .nwb file had contained a Subject group with
    subject_id=0001.  Each file in a subject-specific subfolder follows the
    convention:

        sub-<subject_id>[_key-<value>][_mod1+mod2+...].nwb

    where following keys are considered if present in the data:

        ses -- session_id\n
        tis -- tissue_sample_id\n
        slice -- slice_id\n
        cell -- cell_id\n

    and `modX` are "modalities" as identified based on detected neural data types
    (such as "ecephys", "icephys") per extensions found in nwb-schema definitions:
    https://github.com/NeurodataWithoutBorders/nwb-schema/tree/dev/core

    In addition an "obj" key with a value corresponding to crc32 checksum of
    "object_id" is added if aforementioned keys and the list of modalities are
    not sufficient to disambiguate different files.

    You can visit https://dandiarchive.org for a growing collection of
    (re)organized dandisets.
    """
    from ..dandiset import Dandiset
    from ..metadata import get_metadata
    from ..organize import (
        create_unique_filenames_from_metadata,
        detect_link_type,
        filter_invalid_metadata_rows,
    )
    from ..pynwb_utils import ignore_benign_pynwb_warnings
    from ..utils import Parallel, copy_file, delayed, find_files, load_jsonl, move_file

    in_place = False  # If we deduce that we are organizing in-place

    # will come handy when dry becomes proper separate option
    def dry_print(msg):
        print(f"DRY: {msg}")

    if files_mode == "dry":

        def act(func, *args, **kwargs):
            dry_print(f"{func.__name__} {args}, {kwargs}")

    else:

        def act(func, *args, **kwargs):
            lgr.debug("%s %s %s", func.__name__, args, kwargs)
            return func(*args, **kwargs)

    if dandiset_path is None:
        dandiset = Dandiset.find(os.curdir)
        if not dandiset:
            raise ValueError(
                "No --dandiset-path was provided, and no dandiset was found "
                "in/above current directory"
            )
        dandiset_path = dandiset.path
        del dandiset

    # Early checks to not wait to fail
    if files_mode == "simulate":
        # in this mode we will demand the entire output folder to be absent
        if op.exists(dandiset_path):
            # TODO: RF away
            raise RuntimeError(
                "In simulate mode %r (--top-path) must not exist, we will create it."
                % dandiset_path
            )

    ignore_benign_pynwb_warnings()

    if not paths:
        try:
            Dandiset(dandiset_path)
        except Exception as exc:
            lgr.debug("Failed to find dandiset at %s: %s", dandiset_path, exc)
            raise ValueError(
                f"No dandiset was found at {dandiset_path}, and no "
                f"paths were provided"
            )
        if files_mode not in ("dry", "move"):
            raise ValueError(
                "Only 'dry' or 'move' mode could be used to operate in-place "
                "within a dandiset (no paths were provided)"
            )
        lgr.info(f"We will organize {dandiset_path} in-place")
        in_place = True
        paths = dandiset_path

    if len(paths) == 1 and paths[0].endswith(".json"):
        # Our dumps of metadata
        metadata = load_jsonl(paths[0])
    else:
        paths = list(find_files(r"\.nwb\Z", paths=paths))
        lgr.info("Loading metadata from %d files", len(paths))
        # Done here so we could still reuse cached 'get_metadata'
        # without having two types of invocation and to guard against
        # problematic ones -- we have an explicit option on how to
        # react to those
        # Doesn't play nice with Parallel
        # with tqdm.tqdm(desc="Files", total=len(paths), unit="file", unit_scale=False) as pbar:
        failed = []

        def _get_metadata(path):
            try:
                meta = get_metadata(path)
            except Exception as exc:
                meta = {}
                failed.append(path)
                # pbar.desc = "Files (%d failed)" % len(failed)
                lgr.debug("Failed to get metadata for %s: %s", path, exc)
            # pbar.update(1)
            meta["path"] = path
            return meta

        if not devel_debug:
            # Note: It is Python (pynwb) intensive, not IO, so ATM there is little
            # to no benefit from Parallel without using multiproc!  But that would
            # complicate progress bar indication... TODO
            metadata = list(
                Parallel(n_jobs=-1, verbose=10)(
                    delayed(_get_metadata)(path) for path in paths
                )
            )
        else:
            metadata = list(map(_get_metadata, paths))
        if failed:
            lgr.warning(
                "Failed to load metadata for %d out of %d files",
                len(failed),
                len(paths),
            )

    metadata, skip_invalid = filter_invalid_metadata_rows(metadata)
    if skip_invalid:
        msg = (
            "%d out of %d files were found not containing all necessary "
            "metadata: %s"
            % (
                len(skip_invalid),
                len(metadata) + len(skip_invalid),
                ", ".join(m["path"] for m in skip_invalid),
            )
        )
        if invalid == "fail":
            raise ValueError(msg)
        elif invalid == "warn":
            lgr.warning(msg + " They will be skipped")
        else:
            raise ValueError(f"invalid has an invalid value {invalid}")

    if not op.exists(dandiset_path):
        act(os.makedirs, dandiset_path)

    if files_mode == "auto":
        files_mode = detect_link_type(dandiset_path)

    metadata = create_unique_filenames_from_metadata(metadata)

    # Verify first that the target paths do not exist yet, and fail if they do
    # Note: in "simulate" mode we do early check as well, so this would be
    # duplicate but shouldn't hurt
    existing = []
    for e in metadata:
        dandi_fullpath = op.join(dandiset_path, e["dandi_path"])
        if op.exists(dandi_fullpath):
            # It might be the same file, then we would not complain
            if not (
                op.realpath(e["path"])
                == op.realpath(op.join(dandiset_path, e["dandi_path"]))
            ):
                existing.append(dandi_fullpath)
            # TODO: it might happen that with "move" we are renaming files
            # so there is an existing, which also gets moved away "first"
            # May be we should RF so the actual loop below would be first done
            # "dry", collect info on what is actually to be done, and then we would complain here
    if existing:
        raise AssertionError(
            "%d paths already exist: %s%s.  Remove them first."
            % (
                len(existing),
                ", ".join(existing[:5]),
                " and more" if len(existing) > 5 else "",
            )
        )

    # we should take additional care about paths if both top_path and
    # provided paths are relative
    use_abs_paths = op.isabs(dandiset_path) or any(
        op.isabs(e["path"]) for e in metadata
    )
    skip_same = []
    acted_upon = []
    for e in metadata:
        dandi_path = e["dandi_path"]
        dandi_fullpath = op.join(dandiset_path, dandi_path)
        dandi_abs_fullpath = (
            op.abspath(dandi_fullpath)
            if not op.isabs(dandi_fullpath)
            else dandi_fullpath
        )
        dandi_dirpath = op.dirname(dandi_fullpath)  # could be sub-... subdir

        e_path = e["path"]
        e_abs_path = e_path

        if not op.isabs(e_path):
            e_abs_path = op.abspath(e_path)
            if use_abs_paths:
                e_path = e_abs_path
            elif files_mode == "symlink":  # path should be relative to the target
                e_path = op.relpath(e_abs_path, dandi_dirpath)

        if dandi_abs_fullpath == e_abs_path:
            lgr.debug("Skipping %s since the same in source/destination", e_path)
            skip_same.append(e)
            continue
        elif files_mode == "symlink" and op.realpath(dandi_abs_fullpath) == op.realpath(
            e_abs_path
        ):
            lgr.debug(
                "Skipping %s since mode is symlink and both resolve to the same path",
                e_path,
            )
            skip_same.append(e)
            continue

        if (
            files_mode == "dry"
        ):  # TODO: this is actually a files_mode on top of modes!!!?
            dry_print(f"{e_path} -> {dandi_path}")
        else:
            if not op.exists(dandi_dirpath):
                os.makedirs(dandi_dirpath)
            if files_mode == "simulate":
                os.symlink(e_path, dandi_fullpath)
                continue
            #
            if files_mode == "symlink":
                os.symlink(e_path, dandi_fullpath)
            elif files_mode == "hardlink":
                os.link(e_path, dandi_fullpath)
            elif files_mode == "copy":
                copy_file(e_path, dandi_fullpath)
            elif files_mode == "move":
                move_file(e_path, dandi_fullpath)
            else:
                raise NotImplementedError(files_mode)
            acted_upon.append(e)

    if acted_upon and in_place:
        # We might need to cleanup a bit - e.g. prune empty directories left
        # by the move in in-place mode
        dirs = set(op.dirname(e["path"]) for e in acted_upon)
        for d in sorted(dirs)[::-1]:  # from longest to shortest
            if op.exists(d):
                try:
                    os.rmdir(d)
                    lgr.info(f"Removed empty directory {d}")
                except Exception as exc:
                    lgr.debug("Failed to remove directory %s: %s", d, exc)

    def msg_(msg, n, cond=None):
        if hasattr(n, "__len__"):
            n = len(n)
        if cond is None:
            cond = bool(n)
        if not cond:
            return ""
        return msg % n

    lgr.info(
        "Organized %d%s paths%s.%s Visit %s/",
        len(acted_upon),
        msg_(" out of %d", metadata, len(metadata) != len(acted_upon)),
        msg_(" (%d same existing skipped)", skip_same),
        msg_(" %d invalid not considered.", skip_invalid),
        dandiset_path.rstrip("/"),
    )



from copy import deepcopy
from datetime import datetime
import json
import os
import os.path as op
from pathlib import Path
import re
from uuid import uuid4

import jsonschema

from .dandiset import Dandiset
from . import __version__, get_logger, models
from .pynwb_utils import (
    _get_pynwb_metadata,
    get_neurodata_types,
    get_nwb_version,
    ignore_benign_pynwb_warnings,
    metadata_cache,
)
from .utils import ensure_datetime, get_utcnow_datetime

lgr = get_logger()


@metadata_cache.memoize_path
def get_metadata(path):
    """Get selected metadata from a .nwb file or a dandiset directory

    If a directory given and it is not a Dandiset, None is returned

    Parameters
    ----------
    path: str or Path

    Returns
    -------
    dict
    """
    # when we run in parallel, these annoying warnings appear
    ignore_benign_pynwb_warnings()
    path = str(path)  # for Path
    meta = dict()

    if op.isdir(path):
        try:
            dandiset = Dandiset(path)
            return dandiset.metadata
        except ValueError as exc:
            lgr.debug("Failed to get metadata for %s: %s", path, exc)
            return None

    # First read out possibly available versions of specifications for NWB(:N)
    meta["nwb_version"] = get_nwb_version(path)

    # PyNWB might fail to load because of missing extensions.
    # There is a new initiative of establishing registry of such extensions.
    # Not yet sure if PyNWB is going to provide "native" support for needed
    # functionality: https://github.com/NeurodataWithoutBorders/pynwb/issues/1143
    # So meanwhile, hard-coded workaround for data types we care about
    ndtypes_registry = {
        "AIBS_ecephys": "allensdk.brain_observatory.ecephys.nwb",
        "ndx-labmetadata-abf": "ndx_dandi_icephys",
    }
    tried_imports = set()
    while True:
        try:
            meta.update(_get_pynwb_metadata(path))
            break
        except KeyError as exc:  # ATM there is
            lgr.debug("Failed to read %s: %s", path, exc)
            import re

            res = re.match(r"^['\"\\]+(\S+). not a namespace", str(exc))
            if not res:
                raise
            ndtype = res.groups()[0]
            if ndtype not in ndtypes_registry:
                raise ValueError(
                    "We do not know which extension provides %s. "
                    "Original exception was: %s. " % (ndtype, exc)
                )
            import_mod = ndtypes_registry[ndtype]
            lgr.debug("Importing %r which should provide %r", import_mod, ndtype)
            if import_mod in tried_imports:
                raise RuntimeError(
                    "We already tried importing %s to provide %s, but it seems it didn't help"
                    % (import_mod, ndtype)
                )
            tried_imports.add(import_mod)
            __import__(import_mod)

    meta["nd_types"] = get_neurodata_types(path)

    return meta


def parse_age(age):
    """
    Convert a human-friendly duration string into an ISO 8601 duration

    Parameters
    ----------
    age : str

    Returns
    -------
    str
    """
    m = re.fullmatch(r"(\d+)\s*(y(ear)?|m(onth)?|w(eek)?|d(ay)?)s?", age, flags=re.I)
    if m:
        qty = int(m.group(1))
        unit = m.group(2)[0].upper()
        return f"P{qty}{unit}"
    else:
        raise ValueError(age)


def extract_age(metadata):
    try:
        dob = ensure_datetime(metadata["date_of_birth"])
        start = ensure_datetime(metadata["session_start_time"])
    except (KeyError, TypeError, ValueError):
        try:
            duration = parse_age(metadata["age"])
        except (KeyError, TypeError, ValueError):
            return ...
    else:
        duration = timedelta2duration(start - dob)
    return models.PropertyValue(value=duration, unitText="Years from birth")


def timedelta2duration(delta):
    """
    Convert a datetime.timedelta to ISO 8601 duration format

    Parameters
    ----------
    delta : datetime.timedelta

    Returns
    -------
    str
    """
    s = "P"
    if delta.days:
        s += f"{delta.days}D"
    if delta.seconds or delta.microseconds:
        sec = delta.seconds
        if delta.microseconds:
            # Don't add when microseconds is 0, so that sec will be an int then
            sec += delta.microseconds / 1000000
        s += f"T{sec}S"
    if s == "P":
        s += "0D"
    return s


def extract_sex(metadata):
    value = metadata.get("sex", None)
    if value is not None and value != "":
        value = value.lower()
        if value in ["m", "male"]:
            value_id = "http://purl.obolibrary.org/obo/PATO_0000384"
            value = "Male"
        elif value in ["f", "female"]:
            value_id = "http://purl.obolibrary.org/obo/PATO_0000383"
            value = "Female"
        elif value in ["unknown", "u"]:
            value_id = None
            value = "Unknown"
        elif value in ["other"]:
            value_id = None
            value = "Other"
        elif value.startswith("http"):
            value_id = value
            value = None
        else:
            raise ValueError(f"Cannot interpret sex field: {value}")
        return models.SexType(identifier=value_id, name=value)
    else:
        return ...


def extract_species(metadata):
    value = metadata.get("species", None)
    if value is not None and value != "":
        value = value.lower()
        if "mouse" in value or value.startswith("mus"):
            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_10090"
            value = "House mouse"
        elif "human" in value or value.startswith("homo"):
            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_9606"
            value = "Human"
        elif "norvegicus" in value:
            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_10116"
            value = "Brown rat"
        elif "mulatta" in value or "rhesus" in value:
            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_9544"
            value = "Rhesus monkey"
        elif "jacchus" in value:
            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_9483"
            value = "Common marmoset"
        elif "melanogaster" in value or "fruit fly" in value:
            value_id = "http://purl.obolibrary.org/obo/NCBITaxon_7227"
            value = "Common fruit fly"
        elif value.startswith("http"):
            value_id = value
            value = None
        else:
            raise ValueError(f"Cannot interpret species field: {value}")
        return models.SpeciesType(identifier=value_id, name=value.capitalize())
    else:
        return ...


def extract_assay_type(metadata):
    if "assayType" in metadata:
        return [models.AssayType(identifier="assayType", name=metadata["assayType"])]
    else:
        return ...


def extract_anatomy(metadata):
    if "anatomy" in metadata:
        return [models.Anatomy(identifier="anatomy", name=metadata["anatomy"])]
    else:
        return ...


def extract_model(modelcls, metadata, **kwargs):
    m = modelcls.unvalidated()
    for field in m.__fields__.keys():
        value = kwargs.get(field, extract_field(field, metadata))
        if value is not Ellipsis:
            setattr(m, field, value)
    # return modelcls(**m.dict())
    return m


def extract_model_list(modelcls, id_field, id_source, **kwargs):
    def func(metadata):
        m = extract_model(
            modelcls, metadata, **{id_field: metadata.get(id_source)}, **kwargs
        )
        if all(v is None for k, v in m.dict().items() if k != "schemaKey"):
            return []
        else:
            return [m]

    return func


def extract_wasDerivedFrom(metadata):
    derived_from = None
    for field, sample_name in [
        ("tissue_sample_id", "tissuesample"),
        ("slice_id", "slice"),
        ("cell_id", "cell"),
    ]:
        if metadata.get(field) is not None:
            derived_from = [
                models.BioSample(
                    identifier=metadata[field],
                    wasDerivedFrom=derived_from,
                    sampleType=models.SampleType(name=sample_name),
                )
            ]
    return derived_from


extract_wasAttributedTo = extract_model_list(
    models.Participant, "identifier", "subject_id", id=...
)

extract_wasGeneratedBy = extract_model_list(
    models.Session, "name", "session_id", id=...
)


def extract_digest(metadata):
    if "digest" in metadata:
        return {models.DigestType[metadata["digest_type"]]: metadata["digest"]}
    else:
        return ...


FIELD_EXTRACTORS = {
    "wasDerivedFrom": extract_wasDerivedFrom,
    "wasAttributedTo": extract_wasAttributedTo,
    "wasGeneratedBy": extract_wasGeneratedBy,
    "age": extract_age,
    "sex": extract_sex,
    "assayType": extract_assay_type,
    "anatomy": extract_anatomy,
    "digest": extract_digest,
    "species": extract_species,
}


def extract_field(field, metadata):
    if field in FIELD_EXTRACTORS:
        return FIELD_EXTRACTORS[field](metadata)
    else:
        return metadata.get(field, ...)


def nwb2asset(
    nwb_path, digest=None, digest_type=None, schema_version=None
) -> models.BareAssetMeta:
    if schema_version is not None:
        current_version = models.get_schema_version()
        if schema_version != current_version:
            raise ValueError(
                f"Unsupported schema version: {schema_version}; expected {current_version}"
            )
    start_time = datetime.now().astimezone()
    metadata = get_metadata(nwb_path)
    if digest is not None:
        metadata["digest"] = digest
        metadata["digest_type"] = digest_type
    metadata["contentSize"] = op.getsize(nwb_path)
    metadata["encodingFormat"] = "application/x-nwb"
    metadata["dateModified"] = get_utcnow_datetime()
    metadata["blobDateModified"] = ensure_datetime(os.stat(nwb_path).st_mtime)
    if metadata["blobDateModified"] > metadata["dateModified"]:
        lgr.warning(
            "mtime %s of %s is in the future", metadata["blobDateModified"], nwb_path
        )
    asset = metadata2asset(metadata)
    end_time = datetime.now().astimezone()
    if asset.wasGeneratedBy is None:
        asset.wasGeneratedBy = []
    asset.wasGeneratedBy.append(get_generator(start_time, end_time))
    return asset


def get_default_metadata(path, digest=None, digest_type=None) -> models.BareAssetMeta:
    start_time = datetime.now().astimezone()
    if digest is not None:
        digest_model = {models.DigestType[digest_type]: digest}
    else:
        digest_model = []
    dateModified = get_utcnow_datetime()
    blobDateModified = ensure_datetime(os.stat(path).st_mtime)
    if blobDateModified > dateModified:
        lgr.warning("mtime %s of %s is in the future", blobDateModified, path)
    end_time = datetime.now().astimezone()
    return models.BareAssetMeta.unvalidated(
        contentSize=os.path.getsize(path),
        digest=digest_model,
        dateModified=dateModified,
        blobDateModified=blobDateModified,
        wasGeneratedBy=[get_generator(start_time, end_time)],
        # encodingFormat # TODO
    )


def get_generator(start_time: datetime, end_time: datetime) -> models.Activity:
    return models.Activity(
        id=uuid4().urn,
        name="Metadata generation",
        description="Metadata generated by DANDI cli",
        wasAssociatedWith=[
            models.Software(
                identifier="RRID:SCR_019009",
                name="DANDI Command Line Interface",
                version=__version__,
                url="https://github.com/dandi/dandi-cli",
                schemaKey="Software",
            )
        ],
        startedAt=start_time,
        endedAt=end_time,
    )


def metadata2asset(metadata):
    return extract_model(models.BareAssetMeta, metadata)


"""
The following section converts metadata schema from the current girder dandiset
model to the new schema in dandi-cli. This section should be removed
after the migration is finished to the
"""

mapping = {
    "identifier": ["identifier"],
    "name": ["name"],
    "description": ["description"],
    "contributors": ["contributor"],
    "sponsors": ["contributor", ["Sponsor"]],
    "license": ["license"],
    "keywords": ["keywords"],
    "project": ["generatedBy"],
    "conditions_studied": ["about"],
    "associated_anatomy": ["about"],
    "protocols": ["protocol"],
    "ethicsApprovals": ["ethicsApproval"],
    "access": ["access"],
    "associatedData": ["relatedResource", "IsDerivedFrom"],
    "publications": ["relatedResource", "IsDescribedBy"],
    "age": ["variableMeasured"],
    "organism": ["variableMeasured"],
    "sex": ["variableMeasured"],
    "number_of_subjects": ["assetsSummary", "numberOfSubjects"],
    "number_of_cells": ["assetsSummary", "numberOfCells"],
    "number_of_tissue_samples": ["assetsSummary", "numberOfSamples"],
}


def toContributor(value, contrib_type):
    if not isinstance(value, list):
        value = [value]
    out = []
    for item in value:
        if item == {"orcid": "", "roles": []}:
            continue
        contrib = {}
        if "name" in item:
            name = item["name"].split()
            item["name"] = f"{name[-1]}, {' '.join(name[:-1])}"
        if "roles" in item:
            roles = []
            for role in item["roles"]:
                tmp = role.split()
                if len(tmp) > 1:
                    roles.append("".join([val.capitalize() for val in tmp]))
                else:
                    roles.append(tmp.pop())
            contrib["roleName"] = [getattr(models.RoleType, role) for role in roles]
            del item["roles"]
        elif contrib_type == "sponsors":
            contrib["roleName"] = [models.RoleType.Funder]
        if "awardNumber" in item:
            contrib["awardNumber"] = item["awardNumber"]
            del item["awardNumber"]
        if "orcid" in item:
            if item["orcid"]:
                contrib["identifier"] = item["orcid"]
            # else:
            #    contrib["identifier"] = models.PropertyValue()
            del item["orcid"]
        if "affiliation" in item:
            item["affiliation"] = [models.Organization(name=item["affiliation"])]
        if "affiliations" in item:
            item["affiliation"] = [
                models.Organization(name=affiliate)
                for affiliate in item["affiliations"]
            ]
            del item["affiliations"]
        contrib.update(**{f"{k}": v for k, v in item.items()})
        if "awardNumber" in contrib or contrib_type == "sponsors":
            contrib = models.Organization(**contrib)
        else:
            if "name" not in contrib:
                contrib["name"] = "Last, First"
            contrib = models.Person(**contrib)
        out.append(contrib)
    return out


def convertv1(data):
    oldmeta = deepcopy(data["dandiset"]) if "dandiset" in data else deepcopy(data)
    newmeta = {}
    for oldkey, value in oldmeta.items():
        if oldkey in ["language", "altid", "number_of_slices"]:
            continue
        if oldkey not in mapping:
            raise KeyError(f"Could not find {oldkey}")
        if len(mapping[oldkey]) == 0:
            newkey = f"schema:{oldkey}"
        else:
            newkey = mapping[oldkey][0]
        if oldkey in ["contributors", "sponsors"]:
            value = toContributor(value, oldkey)
        if oldkey == "access":
            value = [
                models.AccessRequirements(
                    status=models.AccessType.Open, email=value["access_contact_email"]
                )
            ]
        if oldkey == "license":
            value = [
                getattr(
                    models.LicenseType,
                    value.replace("dandi", "spdx").replace("-", "_").replace(".", ""),
                )
            ]
        if oldkey == "identifier":
            value = f"DANDI:{value}"
        if len(mapping[oldkey]) == 2:
            extra = mapping[oldkey][1]
            if newkey == "contributor":
                extrakey = "roleName"
            if oldkey == "sponsors":
                extrakey = "roleName"
            if oldkey in ["publications", "associatedData"]:
                extrakey = "relation"
                if not isinstance(value, list):
                    value = [value]
                out = []
                for item in value:
                    if isinstance(item, dict):
                        if (
                            "relation" in item
                            and "publication" in item["relation"].lower()
                        ):
                            del item["relation"]
                        if "relation" not in item:
                            if oldkey == "publications":
                                item["relation"] = models.RelationType.IsDescribedBy
                            if oldkey == "associatedData":
                                item["relation"] = models.RelationType.IsDerivedFrom
                        out.append(models.Resource(**item))
                    elif not any(item in val.dict().values() for val in out):
                        out.append(
                            models.Resource(
                                url=item, relation=models.RelationType.IsDescribedBy
                            )
                        )
                value = out
            if oldkey in [
                "number_of_subjects",
                "number_of_cells",
                "number_of_tissue_samples",
            ]:
                value = {extra: value}
                extrakey = None
            if isinstance(value, list):
                for val in value:
                    if extrakey:
                        if extrakey == "relation":
                            val.relation = getattr(models.RelationType, extra)
                        elif extrakey == "roleName":
                            val.roleName = [
                                getattr(models.RoleType, role) for role in extra
                            ]
                        else:
                            val[extrakey] = extra
            if isinstance(value, dict):
                if extrakey:
                    value[extrakey] = extra
        if newkey == "variableMeasured":
            if oldkey in ["age", "sex"]:
                vm = {"name": oldkey}
                if oldkey == "sex":
                    vm["value"] = value
                else:
                    if "maximum" in value and isinstance(value["maximum"], str):
                        if "days" in value["maximum"]:
                            value["units"] = "days"
                        if "Gestational" in value["maximum"]:
                            value["units"] = "Gestational Week"
                            value["maximum"] = value["maximum"].split()[-1]
                        if value["maximum"].startswith("P"):
                            value["maximum"] = value["maximum"][1:-1]
                            value["units"] = value["maximum"][-1]
                        if "None" not in value["maximum"]:
                            value["maximum"] = float(value["maximum"].split()[0])
                    if "minimum" in value and isinstance(value["minimum"], str):
                        if "days" in value["minimum"]:
                            value["units"] = "days"
                        if "Gestational" in value["minimum"]:
                            value["units"] = "Gestational Week"
                            value["minimum"] = value["minimum"].split()[-1]
                        if value["minimum"].startswith("P"):
                            value["minimum"] = value["minimum"][1:-1]
                            value["units"] = value["minimum"][-1]
                        if "None" not in value["minimum"]:
                            value["minimum"] = float(value["minimum"].split()[0])
                    value["unitText"] = value["units"]
                    del value["units"]
                    vm.update(**value)
            else:
                newvalues = []
                for val in value:
                    if "species" in val:
                        newvalues.append(val["species"])
                vm = {"name": "species", "value": newvalues}
            value = vm
        if newkey == "about":
            newvalues = []
            for item in value:
                id = item.get("identifier", None)
                if id is not None and not id.startswith("http"):
                    item["identifier"] = None
                newvalues.append(models.TypeModel(**item))
            value = newvalues
        if newkey not in newmeta:
            newmeta[newkey] = value
        else:
            curvalue = newmeta[newkey]
            if not isinstance(curvalue, list):
                newmeta[newkey] = [curvalue]
            if not isinstance(value, list):
                value = [value]
            newmeta[newkey].extend(value)
    if "assetsSummary" in newmeta:
        del newmeta["assetsSummary"]
    if "variableMeasured" in newmeta:
        del newmeta["variableMeasured"]
    return newmeta


def migrate2newschema(meta):
    newmeta = convertv1(meta)
    if "version" in newmeta:
        newmeta["id"] = f"{newmeta['identifier']}/{newmeta['version']}"
    else:
        newmeta["id"] = f"{newmeta['identifier']}/draft"
    dandimeta = models.DandisetMeta.unvalidated(**newmeta)
    return dandimeta


def generate_context():
    import pydantic

    fields = {
        "@version": 1.1,
        "dandi": "http://schema.dandiarchive.org/",
        "dandiasset": "http://iri.dandiarchive.org/",
        "DANDI": "http://identifiers.org/DANDI:",
        "dct": "http://purl.org/dc/terms/",
        "owl": "http://www.w3.org/2002/07/owl#",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "rdfa": "http://www.w3.org/ns/rdfa#",
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "schema": "http://schema.org/",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
        "skos": "http://www.w3.org/2004/02/skos/core#",
        "prov": "http://www.w3.org/ns/prov#",
        "pav": "http://purl.org/pav/",
        "nidm": "http://purl.org/nidash/nidm#",
        "uuid": "http://uuid.repronim.org/",
        "rs": "http://schema.repronim.org/",
        "RRID": "https://scicrunch.org/resolver/RRID:",
        "ORCID": "https://orcid.org/",
        "ROR": "https://ror.org/",
        "PATO": "http://purl.obolibrary.org/obo/PATO_",
        "spdx": "http://spdx.org/licenses/",
    }
    for val in dir(models):
        klass = getattr(models, val)
        if not isinstance(klass, pydantic.main.ModelMetaclass):
            continue
        if hasattr(klass, "_ldmeta"):
            if "nskey" in klass._ldmeta:
                name = klass.__name__
                fields[name] = f'{klass._ldmeta["nskey"]}:{name}'
        for name, field in klass.__fields__.items():
            if name == "id":
                fields[name] = "@id"
            elif name == "schemaKey":
                fields[name] = "@type"
            elif name == "digest":
                fields[name] = "@nest"
            elif "nskey" in field.field_info.extra:
                if name not in fields:
                    fields[name] = {"@id": field.field_info.extra["nskey"] + ":" + name}
                    if "List" in str(field.outer_type_):
                        fields[name]["@container"] = "@set"
                    if name == "contributor":
                        fields[name]["@container"] = "@list"
                    if "enum" in str(field.type_) or name == "url":
                        fields[name]["@type"] = "@id"
    for item in models.DigestType:
        fields[item.value] = {"@id": item.value, "@nest": "digest"}
    return {"@context": fields}


def publish_model_schemata(releasedir):
    version = models.get_schema_version()
    vdir = Path(releasedir, version)
    vdir.mkdir(exist_ok=True, parents=True)
    (vdir / "dandiset.json").write_text(models.DandisetMeta.schema_json(indent=2))
    (vdir / "asset.json").write_text(models.AssetMeta.schema_json(indent=2))
    (vdir / "context.json").write_text(json.dumps(generate_context(), indent=2))
    return vdir


def validate_dandiset_json(data, schema_dir):
    with Path(schema_dir, "dandiset.json").open() as fp:
        schema = json.load(fp)
    jsonschema.validate(data, schema)


def validate_asset_json(data, schema_dir):
    with Path(schema_dir, "asset.json").open() as fp:
        schema = json.load(fp)
    jsonschema.validate(data, schema)



import re

import pytest

from ..dandietag import DandiETag, Part, PartGenerator, mb, tb


@pytest.mark.parametrize(
    "file_size,initial_part_size,final_part_size,part_count",
    [
        (0, 0, 0, 0),
        (1, 1, 1, 1),
        (mb(64), mb(64), mb(64), 1),
        (mb(50), mb(50), mb(50), 1),
        (mb(70), mb(64), mb(6), 2),
        (mb(140), mb(64), mb(12), 3),
        (mb(640), mb(64), mb(64), 10),
        (tb(5), 549755814, 549754694, 10000),
    ],
)
def test_part_generator(file_size, initial_part_size, final_part_size, part_count):
    pg = PartGenerator.for_file_size(file_size)
    assert pg == PartGenerator(part_count, initial_part_size, final_part_size)
    assert len(pg) == part_count
    parts = list(pg)
    assert [p.number for p in parts] == list(range(1, part_count + 1))
    assert all(p.size == initial_part_size for p in parts[:-1])
    if parts:
        assert parts[-1].size == final_part_size
    offset = 0
    for p in parts:
        assert p.offset == offset
        offset += p.size
    if parts:
        assert pg[1] == Part(1, 0, initial_part_size)
        assert pg[part_count] == Part(
            part_count, initial_part_size * (part_count - 1), final_part_size
        )
    with pytest.raises(IndexError):
        pg[0]
    with pytest.raises(IndexError):
        pg[part_count + 1]
    with pytest.raises(IndexError):
        pg[-1]


def test_part_generator_too_large():
    with pytest.raises(ValueError) as excinfo:
        PartGenerator.for_file_size(tb(5) + 1)
    assert str(excinfo.value) == "File is larger than the S3 maximum object size."


def test_dandietag_tiny(tmp_path):
    f = tmp_path / "sample.txt"
    f.write_bytes(b"123")
    s = DandiETag.from_file(f).as_str()
    assert s == "d022646351048ac0ba397d12dfafa304-1"
    assert re.fullmatch(DandiETag.REGEX, s)
    assert len(s) <= DandiETag.MAX_STR_LENGTH


def test_dandietag_null(tmp_path):
    f = tmp_path / "sample.dat"
    f.write_bytes(b"\0")
    s = DandiETag.from_file(f).as_str()
    assert s == "7e4696ef25d5faececd853ce5e2a233b-1"
    assert re.fullmatch(DandiETag.REGEX, s)
    assert len(s) <= DandiETag.MAX_STR_LENGTH


PART_DIGESTS = [
    b"\x06\x1c\x9a\xee\xac\x02\x0f\xd8\xa1\xd1\xc9\xcbb\x1d'V",
    b"\xd5z\x92\x92\t\xdd\xfbX\xf6\x05\x83\xcb\xcf\x96\xde%",
    b"z\x88\x9e\xe7m\x9a\n{=\x85,\xc2t_\x1e#",
    b"\x84\xc6<M\x1a$%\xac?\x1a\x0bt\xf0|sY",
    b"\xa3\x98\xc1d\xf5e\xdb;\xd0\xe5\x87\x19\x8d\xcd5\xe9",
    b"\xd7\xe1\x0c\xf2\xfa\x985)\xa8\x8a\xa7\x19z%)\x96",
    b"\xfe\x12\x82\xdb^\x0fr|\x9f\xd5\x924\xba?\xd5D",
    b"*~@\xb4\x19:\xa0\xa2\xc0\xa5q\x82\xcdgv@",
    b"\xaf\x89\x88V^\x00\x9f\x8f\xeb\xf8\x1e\x90\x17\xe4\xf7\xf4",
    b"\xf8\t\xb3fn!\x9c\x13\x8e\x1d\x86\xfd}\x91H\xbf",
]

ETAG = "8e5394f58846c6874be226c5a251f780-10"


def test_add_next_digest():
    etagger = DandiETag(mb(640))
    assert etagger.part_qty == 10
    for i, d in enumerate(PART_DIGESTS):
        assert not etagger.complete
        with pytest.raises(ValueError) as excinfo:
            etagger.as_str()
        assert str(excinfo.value) == "Not all part hashes submitted"
        assert etagger.get_next_part().number == i + 1
        etagger._add_next_digest(d)
    assert etagger.complete
    assert etagger.get_next_part() is None
    s = etagger.as_str()
    assert s == ETAG
    assert re.fullmatch(DandiETag.REGEX, s)
    assert len(s) <= DandiETag.MAX_STR_LENGTH


def test_add_digest_reversed():
    etagger = DandiETag(mb(640))
    assert etagger.part_qty == 10
    assert not etagger.complete
    for p, d in reversed(list(zip(etagger.get_parts(), PART_DIGESTS))):
        assert not etagger.complete
        etagger._add_digest(p, d)
    assert etagger.complete
    assert etagger.as_str() == ETAG


def test_add_digest_out_of_order():
    etagger = DandiETag(mb(640))
    assert etagger.part_qty == 10
    assert not etagger.complete
    pieces = list(zip(etagger.get_parts(), PART_DIGESTS))
    for p, d in pieces[::2] + pieces[1::2]:
        assert not etagger.complete
        etagger._add_digest(p, d)
    assert etagger.complete
    assert etagger.as_str() == ETAG



from datetime import datetime
import logging
import os
from pathlib import Path
import re
import shutil
from subprocess import DEVNULL, check_output, run
import tempfile
from time import sleep
from uuid import uuid4

from dateutil.tz import tzutc
import pynwb
import pytest
import requests

from .skip import skipif
from ..cli.command import organize
from ..consts import dandiset_metadata_file, known_instances
from ..dandiapi import DandiAPIClient
from .. import get_logger
from ..pynwb_utils import make_nwb_file, metadata_nwb_file_fields
from ..upload import upload

lgr = get_logger()


@pytest.fixture(autouse=True)
def capture_all_logs(caplog):
    caplog.set_level(logging.DEBUG, logger="dandi")


# TODO: move into some common fixtures.  We might produce a number of files
#       and also carry some small ones directly in git for regression testing
@pytest.fixture(scope="session")
def simple1_nwb_metadata(tmpdir_factory):
    # very simple assignment with the same values as the key with 1 as suffix
    metadata = {f: "{}1".format(f) for f in metadata_nwb_file_fields}
    metadata["identifier"] = uuid4().hex
    # subject_fields

    # tune specific ones:
    # Needs an explicit time zone since otherwise pynwb would add one
    # But then comparison breaks anyways any ways yoh have tried to set it
    # for datetime.now.  Taking example from pynwb tests
    metadata["session_start_time"] = datetime(2017, 4, 15, 12, tzinfo=tzutc())
    metadata["keywords"] = ["keyword1", "keyword 2"]
    # since NWB 2.1.0 some fields values become "arrays" which are
    # then reloaded as tuples, so we force them being tuples here
    # See e.g. https://github.com/NeurodataWithoutBorders/pynwb/issues/1091
    for f in "related_publications", "experimenter":
        metadata[f] = (metadata[f],)
    return metadata


@pytest.fixture(scope="session")
def simple1_nwb(simple1_nwb_metadata, tmpdir_factory):
    return make_nwb_file(
        str(tmpdir_factory.mktemp("data").join("simple1.nwb")), **simple1_nwb_metadata
    )


@pytest.fixture(scope="session")
def simple2_nwb(simple1_nwb_metadata, tmpdir_factory):
    """With a subject"""
    return make_nwb_file(
        str(tmpdir_factory.mktemp("data").join("simple2.nwb")),
        subject=pynwb.file.Subject(
            subject_id="mouse001",
            date_of_birth=datetime(2016, 12, 1, tzinfo=tzutc()),
            sex="M",
            species="mouse",
        ),
        **simple1_nwb_metadata,
    )


@pytest.fixture(scope="session")
def organized_nwb_dir(simple2_nwb, tmp_path_factory, clirunner):
    tmp_path = tmp_path_factory.mktemp("dandiset")
    (tmp_path / dandiset_metadata_file).write_text("{}\n")
    r = clirunner.invoke(
        organize, ["-f", "copy", "--dandiset-path", str(tmp_path), str(simple2_nwb)]
    )
    assert r.exit_code == 0, r.stdout
    return tmp_path


@pytest.fixture(scope="session")
def organized_nwb_dir2(simple1_nwb_metadata, simple2_nwb, tmp_path_factory, clirunner):
    tmp_path = tmp_path_factory.mktemp("dandiset")

    # need to copy first and then use -f move since we will create one more
    # file to be "organized"
    shutil.copy(str(simple2_nwb), str(tmp_path))
    make_nwb_file(
        str(tmp_path / "simple3.nwb"),
        subject=pynwb.file.Subject(
            subject_id="lizard001",
            date_of_birth=datetime(2016, 12, 1, tzinfo=tzutc()),
            sex="F",
            species="Gekko gecko",
        ),
        **simple1_nwb_metadata,
    )
    (tmp_path / dandiset_metadata_file).write_text("{}\n")
    r = clirunner.invoke(organize, ["-f", "move", "--dandiset-path", str(tmp_path)])
    assert r.exit_code == 0, r.stdout
    assert sum(p.is_dir() for p in tmp_path.iterdir()) == 2
    return tmp_path


@pytest.fixture(scope="session")
def clirunner():
    """A shortcut to get a click.runner to run a command"""
    from click.testing import CliRunner

    yield CliRunner()


def get_gitrepo_fixture(url, committish=None, scope="session"):

    if committish:
        raise NotImplementedError()

    @pytest.fixture(scope=scope)
    def fixture():
        # TODO: adapt reproman.tests.skip collection of skipif conditions
        # skipif.no_network()
        # skipif.no_git()

        path = tempfile.mktemp()  # not using pytest's tmpdir fixture to not
        # collide in different scopes etc. But we
        # would need to remove it ourselves
        lgr.debug("Cloning %r into %r", url, path)
        try:
            runout = run(["git", "clone", url, path])
            if runout.returncode:
                raise RuntimeError(f"Failed to clone {url} into {path}")
            yield path
        finally:
            try:
                shutil.rmtree(path)
            except BaseException as exc:
                lgr.warning("Failed to remove %s - using Windows?: %s", path, exc)

    return fixture


nwb_test_data = get_gitrepo_fixture("http://github.com/dandi-datasets/nwb_test_data")


LOCAL_DOCKER_DIR = Path(__file__).with_name("data") / "dandiarchive-docker"
LOCAL_DOCKER_ENV = LOCAL_DOCKER_DIR.name


@pytest.fixture(scope="session")
def docker_compose_setup():
    skipif.no_network()
    skipif.no_docker_engine()

    # Check that we're running on a Unix-based system (Linux or macOS), as the
    # Docker images don't work on Windows.
    if os.name != "posix":
        pytest.skip("Docker images require Unix host")

    persist = os.environ.get("DANDI_TESTS_PERSIST_DOCKER_COMPOSE")

    create = (
        persist is None
        or run(
            ["docker", "inspect", f"{LOCAL_DOCKER_ENV}_django_1"],
            stdout=DEVNULL,
            stderr=DEVNULL,
        ).returncode
        != 0
    )

    # TODO: Delete this after <https://github.com/dandi/dandi-api/pull/251> is
    # merged:
    env = dict(os.environ)
    env["DJANGO_DANDI_GIRDER_API_URL"] = "http://localhost:8080/api/v1"
    env["DJANGO_DANDI_GIRDER_API_KEY"] = "abc123"

    try:
        if create:
            run(
                ["docker-compose", "run", "--rm", "django", "./manage.py", "migrate"],
                cwd=str(LOCAL_DOCKER_DIR),
                env=env,
                check=True,
            )
            run(
                [
                    "docker-compose",
                    "run",
                    "--rm",
                    "-e",
                    "DJANGO_SUPERUSER_PASSWORD=nsNc48DBiS",
                    "django",
                    "./manage.py",
                    "createsuperuser",
                    "--no-input",
                    "--username",
                    "admin",
                    "--email",
                    "nil@nil.nil",
                ],
                cwd=str(LOCAL_DOCKER_DIR),
                env=env,
                check=True,
            )

        r = check_output(
            [
                "docker-compose",
                "run",
                "--rm",
                "django",
                "./manage.py",
                "drf_create_token",
                "admin",
            ],
            cwd=str(LOCAL_DOCKER_DIR),
            env=env,
            universal_newlines=True,
        )
        m = re.search(r"^Generated token (\w+) for user admin$", r, flags=re.M)
        if not m:
            raise RuntimeError(
                f"Could not extract Django auth token from drf_create_token output: {r!r}"
            )
        django_api_key = m[1]

        if create:
            run(
                ["docker-compose", "up", "-d", "django", "celery", "redirector"],
                cwd=str(LOCAL_DOCKER_DIR),
                env=env,
                check=True,
            )
            API_URL = known_instances["dandi-api-local-docker-tests"].api
            for _ in range(10):
                try:
                    requests.get(f"{API_URL}/dandisets/")
                except requests.ConnectionError:
                    sleep(1)
                else:
                    break
            else:
                raise RuntimeError("Django container did not start up in time")
        yield {"django_api_key": django_api_key}
    finally:
        if persist in (None, "0"):
            run(["docker-compose", "down", "-v"], cwd=str(LOCAL_DOCKER_DIR), check=True)


@pytest.fixture(scope="session")
def local_dandi_api(docker_compose_setup):
    instance_id = "dandi-api-local-docker-tests"
    instance = known_instances[instance_id]
    api_key = docker_compose_setup["django_api_key"]
    client = DandiAPIClient(api_url=instance.api, token=api_key)
    with client.session():
        yield {
            "api_key": api_key,
            "client": client,
            "instance": instance,
            "instance_id": instance_id,
        }


@pytest.fixture()
def text_dandiset(local_dandi_api, monkeypatch, tmp_path_factory):
    dandiset_id = local_dandi_api["client"].create_dandiset("Text Dandiset", {})[
        "identifier"
    ]
    dspath = tmp_path_factory.mktemp("text_dandiset")
    (dspath / dandiset_metadata_file).write_text(f"identifier: '{dandiset_id}'\n")
    (dspath / "file.txt").write_text("This is test text.\n")
    (dspath / "subdir1").mkdir()
    (dspath / "subdir1" / "apple.txt").write_text("Apple\n")
    (dspath / "subdir2").mkdir()
    (dspath / "subdir2" / "banana.txt").write_text("Banana\n")
    (dspath / "subdir2" / "coconut.txt").write_text("Coconut\n")

    def upload_dandiset(paths=None, **kwargs):
        with monkeypatch.context() as m:
            m.setenv("DANDI_API_KEY", local_dandi_api["api_key"])
            upload(
                paths=paths or [],
                dandiset_path=dspath,
                dandi_instance=local_dandi_api["instance_id"],
                devel_debug=True,
                allow_any_path=True,
                validation="skip",
                **kwargs,
            )

    upload_dandiset()
    return {
        "client": local_dandi_api["client"],
        "dspath": dspath,
        "dandiset_id": dandiset_id,
        "reupload": upload_dandiset,
    }



from concurrent.futures import ThreadPoolExecutor, as_completed
from contextlib import contextmanager
import os.path
from pathlib import Path
import re
from threading import Lock
from xml.etree.ElementTree import fromstring

import requests
import tenacity

from .consts import MAX_CHUNK_SIZE, known_instances_rev
from . import get_logger
from .keyring import keyring_lookup
from .utils import USER_AGENT, try_multiple

lgr = get_logger()


# Following class is loosely based on GirderClient, with authentication etc
# being stripped.
# TODO: add copyright/license info
class RESTFullAPIClient(object):
    """A base class for REST clients"""

    def __init__(self, api_url):
        self.api_url = api_url
        self._session = None
        self._headers = {"User-Agent": USER_AGENT}

    @contextmanager
    def session(self, session=None):
        """
        Use a :class:`requests.Session` object for all outgoing requests.
        If `session` isn't passed into the context manager
        then one will be created and yielded. Session objects are useful for enabling
        persistent HTTP connections as well as partially applying arguments to many
        requests, such as headers.

        Note: `session` is closed when the context manager exits, regardless of who
        created it.

        .. code-block:: python

            with client.session() as session:
                session.headers.update({'User-Agent': 'myapp 1.0'})

                for item in items:
                    client.downloadItem(item, fh)

        In the above example, each request will be executed with the User-Agent header
        while reusing the same TCP connection.

        :param session: An existing :class:`requests.Session` object, or None.
        """
        self._session = session if session else requests.Session()

        try:
            yield self._session
        finally:
            # close only if we started a new one
            if not session:
                self._session.close()
            self._session = None

    def _request_func(self, method):
        if self._session is not None:
            return getattr(self._session, method.lower())
        else:
            return getattr(requests, method.lower())

    def send_request(
        self,
        method,
        path,
        parameters=None,
        data=None,
        files=None,
        json=None,
        headers=None,
        json_resp=True,
        retry=None,
        **kwargs,
    ):
        """
        This method looks up the appropriate method, constructs a request URL
        from the base URL, path, and parameters, and then sends the request. If
        the method is unknown or if the path is not found, an exception is
        raised, otherwise a JSON object is returned with the response.

        This is a convenience method to use when making basic requests that do
        not involve multipart file data that might need to be specially encoded
        or handled differently.

        :param method: The HTTP method to use in the request (GET, POST, etc.)
        :type method: str
        :param path: A string containing the path elements for this request.
            Note that the path string should not begin or end with the path  separator, '/'.
        :type path: str
        :param parameters: A dictionary mapping strings to strings, to be used
            as the key/value pairs in the request parameters.
        :type parameters: dict
        :param data: A dictionary, bytes or file-like object to send in the body.
        :param files: A dictionary of 'name' => file-like-objects for multipart encoding upload.
        :type files: dict
        :param json: A JSON object to send in the request body.
        :type json: dict
        :param headers: If present, a dictionary of headers to encode in the request.
        :type headers: dict
        :param json_resp: Whether the response should be parsed as JSON. If False, the raw
            response object is returned. To get the raw binary content of the response,
            use the ``content`` attribute of the return value, e.g.

            .. code-block:: python

                resp = client.get('my/endpoint', json_resp=False)
                print(resp.content)  # Raw binary content
                print(resp.headers)  # Dict of headers

        :type json_resp: bool
        :param retry: an optional tenacity `retry` argument for retrying the
            request method
        """
        if not parameters:
            parameters = {}

        # Look up the HTTP method we need
        f = self._request_func(method)

        url = self.get_url(path)

        # Make the request, passing parameters and authentication info
        _headers = dict(self._headers)
        if headers:
            _headers.update(headers)

        if json_resp and "accept" not in _headers:
            _headers["accept"] = "application/json"

        lgr.debug("%s %s", method.upper(), url)

        # urllib3's ConnectionPool isn't thread-safe, so we sometimes hit
        # ConnectionErrors on the start of an upload.  Retry when this happens.
        # Cf. <https://github.com/urllib3/urllib3/issues/951>.
        doretry = tenacity.retry_if_exception_type(
            requests.ConnectionError
        ) | tenacity.retry_if_result(lambda r: r.status_code == 503)
        if retry is not None:
            doretry |= retry

        try:
            result = try_multiple(5, doretry, 1.1)(
                f,
                url,
                params=parameters,
                data=data,
                files=files,
                json=json,
                headers=_headers,
                **kwargs,
            )
        except Exception:
            lgr.exception("HTTP connection failed")
            raise

        lgr.debug("Response: %d", result.status_code)

        # If success, return the json object. Otherwise throw an exception.
        if not result.ok:
            msg = f"Error {result.status_code} while sending {method} request to {url}"
            if result.status_code == 409:
                # Blob exists on server; log at DEBUG level
                lgr.debug("%s: %s", msg, result.text)
            else:
                lgr.error("%s: %s", msg, result.text)
            raise requests.HTTPError(msg, response=result)

        if json_resp:
            if result.text.strip():
                return result.json()
            else:
                return None
        else:
            return result

    def get_url(self, path):
        # Construct the url
        if path.lower().startswith(("http://", "https://")):
            return path
        else:
            return self.api_url.rstrip("/") + "/" + path.lstrip("/")

    def get(self, path, parameters=None, json_resp=True):
        """
        Convenience method to call :py:func:`send_request` with the 'GET' HTTP method.
        """
        return self.send_request("GET", path, parameters, json_resp=json_resp)

    def post(
        self,
        path,
        parameters=None,
        files=None,
        data=None,
        json=None,
        headers=None,
        json_resp=True,
    ):
        """
        Convenience method to call :py:func:`send_request` with the 'POST' HTTP method.
        """
        return self.send_request(
            "POST",
            path,
            parameters,
            files=files,
            data=data,
            json=json,
            headers=headers,
            json_resp=json_resp,
        )

    def put(
        self, path, parameters=None, data=None, json=None, json_resp=True, retry=None
    ):
        """
        Convenience method to call :py:func:`send_request` with the 'PUT'
        HTTP method.
        """
        return self.send_request(
            "PUT",
            path,
            parameters,
            data=data,
            json=json,
            json_resp=json_resp,
            retry=retry,
        )

    def delete(self, path, parameters=None, json_resp=True):
        """
        Convenience method to call :py:func:`send_request` with the 'DELETE' HTTP method.
        """
        return self.send_request("DELETE", path, parameters, json_resp=json_resp)

    def patch(self, path, parameters=None, data=None, json=None, json_resp=True):
        """
        Convenience method to call :py:func:`send_request` with the 'PATCH' HTTP method.
        """
        return self.send_request(
            "PATCH", path, parameters, data=data, json=json, json_resp=json_resp
        )


class DandiAPIClient(RESTFullAPIClient):
    def __init__(self, api_url, token=None):
        super().__init__(api_url)
        if token is not None:
            self.authenticate(token)

    def authenticate(self, token):
        self._headers["Authorization"] = f"token {token}"

    def dandi_authenticate(self):
        # Shortcut for advanced folks
        api_key = os.environ.get("DANDI_API_KEY", None)
        if api_key:
            self.authenticate(api_key)
            return
        if self.api_url in known_instances_rev:
            client_name = known_instances_rev[self.api_url]
        else:
            raise NotImplementedError("TODO client name derivation for keyring")
        app_id = f"dandi-api-{client_name}"
        keyring_backend, api_key = keyring_lookup(app_id, "key")
        if not api_key:
            api_key = input(f"Please provide API Key for {client_name}: ")
            keyring_backend.set_password(app_id, "key", api_key)
            lgr.debug("Stored key in keyring")
        self.authenticate(api_key)

    def get_asset(self, dandiset_id, version, asset_id):
        """

        /dandisets/{version__dandiset__pk}/versions/{version__version}/assets/{asset_id}/

        Parameters
        ----------
        dandiset_id
        version
        asset_id

        Returns
        -------

        """
        return self.get(
            f"/dandisets/{dandiset_id}/versions/{version}/assets/{asset_id}/"
        )

    def get_dandiset(self, dandiset_id, version):
        return self._migrate_dandiset_metadata(
            self.get(f"/dandisets/{dandiset_id}/versions/{version}/info/")
        )

    def set_dandiset_metadata(self, dandiset_id, *, metadata):
        # CLI should not update metadata for released dandisets so always "draft"
        return self.put(
            f"/dandisets/{dandiset_id}/versions/draft/",
            json={"metadata": metadata, "name": metadata.get("name", "")},
        )

    def delete_dandiset(self, dandiset_id):
        self.delete(f"/dandisets/{dandiset_id}/")

    def get_dandiset_assets(
        self, dandiset_id, version, page_size=None, path=None, include_metadata=False
    ):
        """ A generator to provide asset records """
        resp = self.get(
            f"/dandisets/{dandiset_id}/versions/{version}/assets/",
            parameters={"page_size": page_size, "path": path},
        )
        while True:
            for asset in resp["results"]:
                if include_metadata:
                    asset["metadata"] = self.get_asset(
                        dandiset_id, version, asset["asset_id"]
                    )
                yield asset
            if resp.get("next"):
                resp = self.get(resp["next"])
            else:
                break

    def get_dandiset_and_assets(self, dandiset_id, version, include_metadata=False):
        """This is pretty much an adapter to provide "harmonized" output in both
        girder and DANDI api clients.

        Harmonization should happen toward DANDI API BUT AFAIK it is still influx
        """
        lgr.info(f"Traversing {dandiset_id} (version: {version})")
        dandiset = self.get_dandiset(dandiset_id, version)
        assets = self.get_dandiset_assets(
            dandiset_id, version, include_metadata=include_metadata
        )
        return dandiset, assets

    def get_download_file_iter(
        self, dandiset_id, version, asset_id, chunk_size=MAX_CHUNK_SIZE
    ):
        url = self.get_url(
            f"/dandisets/{dandiset_id}/versions/{version}/assets/{asset_id}/download/"
        )

        def downloader(start_at=0):
            lgr.debug("Starting download from %s", url)
            headers = None
            if start_at > 0:
                headers = {"Range": f"bytes={start_at}-"}
            result = (self._session if self._session else requests).get(
                url, stream=True, headers=headers
            )
            # TODO: apparently we might need retries here as well etc
            # if result.status_code not in (200, 201):
            result.raise_for_status()

            for chunk in result.iter_content(chunk_size=chunk_size):
                if chunk:  # could be some "keep alive"?
                    yield chunk
            lgr.info("Asset %s successfully downloaded", asset_id)

        return downloader

    # TODO: remove when API stabilizes

    # Should perform changes in-place but also return the original record

    @classmethod
    def _migrate_dandiset_metadata(cls, dandiset):
        dandiset_metadata = dandiset.get("metadata", {})
        if not dandiset_metadata:
            return dandiset
        # DANDI API has no versioning yet, and things are in flux.
        # It used to have metadata within a key... just in case let's also
        # be able to handle "old" style
        if "identifier" not in dandiset_metadata and "dandiset" in dandiset_metadata:
            dandiset["metadata"] = dandiset_metadata.pop("dandiset")
        return dandiset

    def upload(self, dandiset_id, version_id, asset_metadata, filepath, jobs=None):
        """
        Parameters
        ----------
        dandiset_id: str
          the ID of the Dandiset to which to upload the file
        version_id: str
          the ID of the version of the Dandiset to which to upload the file
        asset_metadata: dict
          Metadata for the uploaded asset file.  Must include a "path" field
          giving the POSIX path at which the uploaded file will be placed on
          the server.
        filepath: str or PathLike
          the path to the local file to upload
        """
        for _ in self.iter_upload(
            dandiset_id, version_id, asset_metadata, filepath, jobs=jobs
        ):
            pass

    def iter_upload(self, dandiset_id, version_id, asset_metadata, filepath, jobs=None):
        """
        Parameters
        ----------
        dandiset_id: str
          the ID of the Dandiset to which to upload the file
        version_id: str
          the ID of the version of the Dandiset to which to upload the file
        asset_metadata: dict
          Metadata for the uploaded asset file.  Must include a "path" field
          giving the POSIX path at which the uploaded file will be placed on
          the server.
        filepath: str or PathLike
          the path to the local file to upload

        Returns
        -------
        a generator of `dict`s containing at least a ``"status"`` key
        """
        from .support.digests import get_dandietag

        asset_path = asset_metadata["path"]
        yield {"status": "calculating etag"}
        etagger = get_dandietag(filepath)
        filetag = etagger.as_str()
        lgr.debug("Calculated dandi-etag of %s for %s", filetag, filepath)
        digest = asset_metadata.get("digest", {})
        if "dandi:dandi-etag" in digest:
            if digest["dandi:dandi-etag"] != filetag:
                raise RuntimeError(
                    f"{filepath}: File etag changed; was originally"
                    f" {digest['dandi:dandi-etag']} but is now {filetag}"
                )
        yield {"status": "initiating upload"}
        lgr.debug("%s: Beginning upload", asset_path)
        total_size = os.path.getsize(filepath)
        try:
            resp = self.post(
                "/uploads/initialize/",
                json={
                    "contentSize": total_size,
                    "digest": {
                        "algorithm": "dandi:dandi-etag",
                        "value": filetag,
                    },
                },
            )
        except requests.HTTPError as e:
            if e.response.status_code == 409:
                lgr.debug("%s: Blob already exists on server", asset_path)
                blob_id = e.response.headers["Location"]
            else:
                raise
        else:
            upload_id = resp["upload_id"]
            parts = resp["parts"]
            if len(parts) != etagger.part_qty:
                raise RuntimeError(
                    f"Server and client disagree on number of parts for upload;"
                    f" server says {len(parts)}, client says {etagger.part_qty}"
                )
            parts_out = []
            bytes_uploaded = 0
            storage = RESTFullAPIClient("http://nil.nil")
            lgr.debug("Uploading %s in %d parts", filepath, len(parts))
            with storage.session():
                with open(filepath, "rb") as fp:
                    with ThreadPoolExecutor(max_workers=jobs or 5) as executor:
                        lock = Lock()
                        futures = [
                            executor.submit(
                                upload_part,
                                storage_session=storage,
                                fp=fp,
                                lock=lock,
                                etagger=etagger,
                                asset_path=asset_path,
                                part=part,
                            )
                            for part in parts
                        ]
                        for fut in as_completed(futures):
                            out_part = fut.result()
                            bytes_uploaded += out_part["size"]
                            yield {
                                "status": "uploading",
                                "upload": 100 * bytes_uploaded / total_size,
                                "current": bytes_uploaded,
                            }
                            parts_out.append(out_part)
                lgr.debug("%s: Completing upload", asset_path)
                resp = self.post(
                    f"/uploads/{upload_id}/complete/",
                    json={"parts": parts_out},
                )
                lgr.debug(
                    "%s: Announcing completion to %s",
                    asset_path,
                    resp["complete_url"],
                )
                r = storage.post(
                    resp["complete_url"], data=resp["body"], json_resp=False
                )
                lgr.debug(
                    "%s: Upload completed. Response content: %s",
                    asset_path,
                    r.content,
                )
                rxml = fromstring(r.text)
                m = re.match(r"\{.+?\}", rxml.tag)
                ns = m.group(0) if m else ""
                final_etag = rxml.findtext(f"{ns}ETag")
                if final_etag is not None:
                    final_etag = final_etag.strip('"')
                    if final_etag != filetag:
                        raise RuntimeError(
                            "Server and client disagree on final ETag of uploaded file;"
                            f" server says {final_etag}, client says {filetag}"
                        )
                # else: Error? Warning?
                resp = self.post(f"/uploads/{upload_id}/validate/")
                blob_id = resp["blob_id"]
        lgr.debug("%s: Assigning asset blob to dandiset & version", asset_path)
        yield {"status": "producing asset"}
        extant = self.get_asset_bypath(dandiset_id, version_id, asset_path)
        if extant is None:
            self.post(
                f"/dandisets/{dandiset_id}/versions/{version_id}/assets/",
                json={"metadata": asset_metadata, "blob_id": blob_id},
            )
        else:
            lgr.debug("%s: Asset already exists at path; updating", asset_path)
            self.put(
                f"/dandisets/{dandiset_id}/versions/{version_id}/assets/{extant['asset_id']}/",
                json={"metadata": asset_metadata, "blob_id": blob_id},
            )
        lgr.info("%s: Asset successfully uploaded", asset_path)
        yield {"status": "done"}

    def create_dandiset(self, name, metadata):
        return self.post("/dandisets/", json={"name": name, "metadata": metadata})

    def download_asset(
        self, dandiset_id, version, asset_id, filepath, chunk_size=MAX_CHUNK_SIZE
    ):
        downloader = self.get_download_file_iter(
            dandiset_id, version, asset_id, chunk_size=chunk_size
        )
        with open(filepath, "wb") as fp:
            for chunk in downloader():
                fp.write(chunk)

    def download_asset_bypath(
        self, dandiset_id, version, asset_path, filepath, chunk_size=MAX_CHUNK_SIZE
    ):
        asset = self.get_asset_bypath(dandiset_id, version, asset_path)
        if asset is None:
            raise RuntimeError(f"No asset found with path {asset_path!r}")
        self.download_asset(
            dandiset_id, version, asset["asset_id"], filepath, chunk_size=chunk_size
        )

    def download_assets_directory(
        self, dandiset_id, version, assets_dirpath, dirpath, chunk_size=MAX_CHUNK_SIZE
    ):
        if assets_dirpath and not assets_dirpath.endswith("/"):
            assets_dirpath += "/"
        assets = list(
            self.get_dandiset_assets(dandiset_id, version, path=assets_dirpath)
        )
        for a in assets:
            filepath = Path(dirpath, a["path"][len(assets_dirpath) :])
            filepath.parent.mkdir(parents=True, exist_ok=True)
            self.download_asset(
                dandiset_id, version, a["asset_id"], filepath, chunk_size=chunk_size
            )

    def get_asset_bypath(
        self, dandiset_id, version, asset_path, include_metadata=False
    ):
        try:
            # Weed out any assets that happen to have the given path as a
            # proper prefix:
            (asset,) = (
                a
                for a in self.get_dandiset_assets(
                    dandiset_id,
                    version,
                    path=asset_path,
                    include_metadata=include_metadata,
                )
                if a["path"] == asset_path
            )
        except ValueError:
            return None
        else:
            return asset

    def publish_version(self, dandiset_id, base_version_id):
        return self.post(
            f"/dandisets/{dandiset_id}/versions/{base_version_id}/publish/"
        )

    def delete_asset(self, dandiset_id, version_id, asset_id):
        self.delete(
            f"/dandisets/{dandiset_id}/versions/{version_id}/assets/{asset_id}/"
        )

    def delete_asset_bypath(self, dandiset_id, version_id, asset_path):
        asset = self.get_asset_bypath(dandiset_id, version_id, asset_path)
        if asset is None:
            raise RuntimeError(f"No asset found with path {asset_path!r}")
        self.delete_asset(dandiset_id, version_id, asset["asset_id"])


def upload_part(storage_session, fp, lock, etagger, asset_path, part):
    etag_part = etagger.get_part(part["part_number"])
    if part["size"] != etag_part.size:
        raise RuntimeError(
            f"Server and client disagree on size of upload part"
            f" {part['part_number']}; server says {part['size']},"
            f" client says {etag_part.size}"
        )
    with lock:
        fp.seek(etag_part.offset)
        chunk = fp.read(part["size"])
    if len(chunk) != part["size"]:
        raise RuntimeError(
            f"End of file {fp.name} reached unexpectedly early:"
            f" read {len(chunk)} bytes of out of an expected {part['size']}"
        )
    lgr.debug(
        "%s: Uploading part %d/%d (%d bytes)",
        asset_path,
        part["part_number"],
        etagger.part_qty,
        part["size"],
    )
    r = storage_session.put(
        part["upload_url"],
        data=chunk,
        json_resp=False,
        retry=tenacity.retry_if_result(lambda r: r.status_code == 500),
    )
    server_etag = r.headers["ETag"].strip('"')
    lgr.debug(
        "%s: Part upload finished ETag=%s Content-Length=%s",
        asset_path,
        server_etag,
        r.headers.get("Content-Length"),
    )
    client_etag = etagger.get_part_etag(etag_part)
    if server_etag != client_etag:
        raise RuntimeError(
            f"Server and client disagree on ETag of upload part"
            f" {part['part_number']}; server says"
            f" {server_etag}, client says {client_etag}"
        )
    return {
        "part_number": part["part_number"],
        "size": part["size"],
        "etag": server_etag,
    }



import os.path
from pathlib import Path
import random
from shutil import rmtree

from ..consts import dandiset_metadata_file
from ..download import download
from ..upload import upload
from ..utils import find_files


def test_upload(local_dandi_api, simple1_nwb, tmp_path):
    client = local_dandi_api["client"]
    r = client.create_dandiset(name="Upload Test", metadata={})
    dandiset_id = r["identifier"]
    client.upload(dandiset_id, "draft", {"path": "testing/simple1.nwb"}, simple1_nwb)
    (asset,) = client.get_dandiset_assets(dandiset_id, "draft")
    assert asset["path"] == "testing/simple1.nwb"
    client.download_assets_directory(dandiset_id, "draft", "", tmp_path)
    (p,) = [p for p in tmp_path.glob("**/*") if p.is_file()]
    assert p == tmp_path / "testing" / "simple1.nwb"
    assert p.stat().st_size == os.path.getsize(simple1_nwb)


def test_publish_and_manipulate(local_dandi_api, monkeypatch, tmp_path):
    client = local_dandi_api["client"]
    dandiset_id = client.create_dandiset("Test Dandiset", {})["identifier"]
    upload_dir = tmp_path / "upload"
    upload_dir.mkdir()
    (upload_dir / dandiset_metadata_file).write_text(f"identifier: '{dandiset_id}'\n")
    (upload_dir / "subdir").mkdir()
    (upload_dir / "subdir" / "file.txt").write_text("This is test text.\n")
    monkeypatch.chdir(upload_dir)
    monkeypatch.setenv("DANDI_API_KEY", local_dandi_api["api_key"])
    upload(
        paths=[],
        dandi_instance=local_dandi_api["instance_id"],
        devel_debug=True,
        allow_any_path=True,
        validation="skip",
    )

    version_id = client.publish_version(dandiset_id, "draft")["version"]

    download_dir = tmp_path / "download"
    download_dir.mkdir()

    def downloaded_files():
        return list(map(Path, find_files(r".*", paths=[download_dir])))

    dandiset_yaml = download_dir / dandiset_id / dandiset_metadata_file
    file_in_version = download_dir / dandiset_id / "subdir" / "file.txt"

    download(
        f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}/versions/{version_id}",
        download_dir,
    )
    assert downloaded_files() == [dandiset_yaml, file_in_version]
    assert file_in_version.read_text() == "This is test text.\n"

    (upload_dir / "subdir" / "file.txt").write_text("This is different text.\n")
    upload(
        paths=[],
        dandi_instance=local_dandi_api["instance_id"],
        devel_debug=True,
        allow_any_path=True,
        validation="skip",
    )
    rmtree(download_dir / dandiset_id)
    download(
        f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}/versions/{version_id}",
        download_dir,
    )
    assert downloaded_files() == [dandiset_yaml, file_in_version]
    assert file_in_version.read_text() == "This is test text.\n"

    (upload_dir / "subdir" / "file2.txt").write_text("This is more text.\n")
    upload(
        paths=[],
        dandi_instance=local_dandi_api["instance_id"],
        devel_debug=True,
        allow_any_path=True,
        validation="skip",
    )

    rmtree(download_dir / dandiset_id)
    download(
        f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}/versions/draft",
        download_dir,
    )
    assert sorted(downloaded_files()) == [
        dandiset_yaml,
        file_in_version,
        file_in_version.with_name("file2.txt"),
    ]
    assert file_in_version.read_text() == "This is different text.\n"
    assert file_in_version.with_name("file2.txt").read_text() == "This is more text.\n"

    rmtree(download_dir / dandiset_id)
    download(
        f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}/versions/{version_id}",
        download_dir,
    )
    assert downloaded_files() == [dandiset_yaml, file_in_version]
    assert file_in_version.read_text() == "This is test text.\n"

    client.delete_asset_bypath(dandiset_id, "draft", "subdir/file.txt")

    rmtree(download_dir / dandiset_id)
    download(
        f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}/versions/draft",
        download_dir,
    )
    assert downloaded_files() == [dandiset_yaml, file_in_version.with_name("file2.txt")]
    assert file_in_version.with_name("file2.txt").read_text() == "This is more text.\n"

    rmtree(download_dir / dandiset_id)
    download(
        f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}/versions/{version_id}",
        download_dir,
    )
    assert downloaded_files() == [dandiset_yaml, file_in_version]
    assert file_in_version.read_text() == "This is test text.\n"


def test_get_asset_include_metadata(local_dandi_api, simple1_nwb, tmp_path):
    client = local_dandi_api["client"]
    r = client.create_dandiset(name="Include Metadata Test", metadata={})
    dandiset_id = r["identifier"]
    client.upload(
        dandiset_id,
        "draft",
        {"path": "testing/simple1.nwb", "foo": "bar"},
        simple1_nwb,
    )

    (asset,) = client.get_dandiset_assets(dandiset_id, "draft")
    assert "metadata" not in asset
    (asset,) = client.get_dandiset_assets(dandiset_id, "draft", include_metadata=True)
    assert asset["metadata"]["path"] == "testing/simple1.nwb"
    assert asset["metadata"]["foo"] == "bar"

    _, (asset,) = client.get_dandiset_and_assets(dandiset_id, "draft")
    assert "metadata" not in asset
    _, (asset,) = client.get_dandiset_and_assets(
        dandiset_id, "draft", include_metadata=True
    )
    assert asset["metadata"]["path"] == "testing/simple1.nwb"
    assert asset["metadata"]["foo"] == "bar"

    asset = client.get_asset_bypath(dandiset_id, "draft", "testing/simple1.nwb")
    assert asset is not None
    assert "metadata" not in asset
    asset = client.get_asset_bypath(
        dandiset_id, "draft", "testing/simple1.nwb", include_metadata=True
    )
    assert asset is not None
    assert asset["metadata"]["path"] == "testing/simple1.nwb"
    assert asset["metadata"]["foo"] == "bar"


def test_large_upload(local_dandi_api, tmp_path):
    client = local_dandi_api["client"]
    asset_file = tmp_path / "asset.dat"
    meg = bytes(random.choices(range(256), k=1 << 20))
    with asset_file.open("wb") as fp:
        for _ in range(100):
            fp.write(meg)
    r = client.create_dandiset(name="Large Upload Test", metadata={})
    dandiset_id = r["identifier"]
    client.upload(dandiset_id, "draft", {"path": "testing/asset.dat"}, asset_file)



"""Various helpful iterators"""

from queue import Empty, Queue
from threading import Thread


class IteratorWithAggregation:
    """
    An iterable over an iterable which also makes an aggregate of the values available asap

    It iterates over the iterable in a separate thread.

    A use case is a generator which collects information about resources,
    which might be relatively fast but still take time.  While we are iterating over it,
    we could perform other operations on yielded records, but we would also like to have access to
    the "summary" object as soon as that iterator completes but while we might still be
    iterating over items in the outside loop.

    Use case: iterate over remote resource for downloads, and get "Total" size/number as
    soon as it becomes known inside the underlying iterator.

    TODO: probably could be more elegant etc if implemented via async/coroutines.

    Attributes
    ----------
    .total:
      Aggregated value as known to the moment. None if nothing was aggregated.
      It is a final value if `finished` is True.
    .finished: bool
      Set to True upon completion of iteration
    .exc: BaseException or None
      If not None -- the exception which was raised

    Example
    -------

    Very simplistic example, since typically (not range) it would be taking some time to
    iterate for the nested iteration::

        it = IteratorWithAggregation(range(3), lambda v, t=0: v+t)
        for v in it:
            print(it.total, it.finished, v)
            sleep(0.02)  # doing smth heavy, but we would know .total as soon as it is known

    would produce (so 3 is known right away, again since it is just range)

        3 True 0
        3 True 1
        3 True 2

    """

    def __init__(self, gen, agg, reraise_immediately=False):
        """

        Parameters
        ----------
        gen: iterable
          Generator (but could be any iterable, but it would not make much sense)
          to yield from
        agg: callable
          A callable with two args: new_value[, total=None] which should return adjusted
          total. Upon first iteration, no prior `total` is provided
        reraise_immediately: bool, optional
          If True, it would stop yielding values as soon as it detects that some
          exception has occurred (although there might still be values in the queue to be yielded
          which were collected before the exception was raised)
        """
        self.gen = gen
        self.agg = agg
        self.reraise_immediately = reraise_immediately

        self.total = None
        self.finished = None
        self._exc = None

    def __iter__(self):
        self.finished = False
        self._exc = None

        queue = Queue()

        def worker():
            """That is the one which interrogates gen and places total
            into queue_total upon completion"""
            total = None
            try:
                for value in self.gen:
                    queue.put(value)
                    self.total = total = (
                        self.agg(value, total) if total is not None else self.agg(value)
                    )
            except BaseException as e:
                self._exc = e
            finally:
                self.finished = True

        t = Thread(target=worker)
        t.start()

        # yield from the queue (.total and .finished could be accessed meanwhile)
        while True:
            if self.reraise_immediately and self._exc is not None:
                break

            # race condition HERE between checking for self.finished and
            if self.finished and queue.empty():
                break
            # in general queue should not be empty, but if it is, e.g. due to race
            # condition with above check
            try:
                yield queue.get(timeout=0.001)
            except Empty:
                continue
        t.join()
        if self._exc is not None:
            raise self._exc



import os.path as op

from .consts import dandiset_metadata_file
from . import get_logger
from .metadata import get_metadata
from .pynwb_utils import validate as pynwb_validate
from .pynwb_utils import validate_cache
from .utils import find_dandi_files, yaml_load

lgr = get_logger()

# TODO -- should come from schema.  This is just a simplistic example for now
_required_dandiset_metadata_fields = ["identifier", "name", "description"]
_required_nwb_metadata_fields = ["subject_id"]


# TODO: provide our own "errors" records, which would also include warnings etc
def validate(paths, schema_version=None, devel_debug=False):
    """Validate content

    Parameters
    ----------
    paths: str or list of paths
      Could be individual (.nwb) files or a single dandiset path.

    Yields
    ------
    path, errors
      errors for a path
    """
    for path in find_dandi_files(paths):
        errors = validate_file(
            path, schema_version=schema_version, devel_debug=devel_debug
        )
        yield path, errors


def validate_file(filepath, schema_version=None, devel_debug=False):
    if op.basename(filepath) == dandiset_metadata_file:
        return validate_dandiset_yaml(
            filepath, schema_version=None, devel_debug=devel_debug
        )
    else:
        return pynwb_validate(filepath, devel_debug=devel_debug) + validate_dandi_nwb(
            filepath, schema_version=schema_version, devel_debug=devel_debug
        )


@validate_cache.memoize_path
def validate_dandiset_yaml(filepath, schema_version=None, devel_debug=False):
    """Validate dandiset.yaml"""
    with open(filepath) as f:
        meta = yaml_load(f, typ="safe")
    if schema_version is None:
        schema_version = meta.get("schemaVersion")
    if schema_version is None:
        return _check_required_fields(meta, _required_dandiset_metadata_fields)
    else:
        from pydantic import ValidationError

        from .metadata import migrate2newschema
        from .models import DandisetMeta, get_schema_version

        current_version = get_schema_version()
        if schema_version != current_version:
            raise ValueError(
                f"Unsupported schema version: {schema_version}; expected {current_version}"
            )
        try:
            new_meta = migrate2newschema(meta)
            DandisetMeta(**new_meta.dict())
        except ValidationError as e:
            if devel_debug:
                raise
            lgr.warning(
                "Validation error for %s: %s", filepath, e, extra={"validating": True}
            )
            return [str(e)]
        except Exception as e:
            if devel_debug:
                raise
            lgr.warning(
                "Unexpected validation error for %s: %s",
                filepath,
                e,
                extra={"validating": True},
            )
            return [f"Failed to convert metadata: {e}"]
        return []


@validate_cache.memoize_path
def validate_dandi_nwb(filepath, schema_version=None, devel_debug=False):
    """Provide validation of .nwb file regarding requirements we impose"""
    if schema_version is not None:
        from pydantic import ValidationError

        from .metadata import nwb2asset
        from .models import BareAssetMeta, get_schema_version

        current_version = get_schema_version()
        if schema_version != current_version:
            raise ValueError(
                f"Unsupported schema version: {schema_version}; expected {current_version}"
            )
        try:
            asset = nwb2asset(filepath, digest="dummy_value", digest_type="sha1")
            BareAssetMeta(**asset.dict())
        except ValidationError as e:
            if devel_debug:
                raise
            lgr.warning(
                "Validation error for %s: %s", filepath, e, extra={"validating": True}
            )
            return [str(e)]
        except Exception as e:
            if devel_debug:
                raise
            lgr.warning(
                "Unexpected validation error for %s: %s",
                filepath,
                e,
                extra={"validating": True},
            )
            return [f"Failed to read metadata: {e}"]
        return []
    else:
        # make sure that we have some basic metadata fields we require
        try:
            meta = get_metadata(filepath)
        except Exception as e:
            if devel_debug:
                raise
            lgr.warning(
                "Failed to read metadata in %s: %s",
                filepath,
                e,
                extra={"validating": True},
            )
            return [f"Failed to read metadata: {e}"]
        return _check_required_fields(meta, _required_nwb_metadata_fields)


def _check_required_fields(d, required):
    errors = []
    for f in required:
        v = d.get(f, None)
        if not v or (isinstance(v, str) and not (v.strip())):
            errors += [f"Required field {f!r} has no value"]
        if v in ("REQUIRED", "PLACEHOLDER"):
            errors += [f"Required field {f!r} has value {v!r}"]
    return errors



import click

from .base import (
    devel_debug_option,
    devel_option,
    instance_option,
    map_to_click_exceptions,
)


class IntColonInt(click.ParamType):
    name = "int:int"

    def convert(self, value, param, ctx):
        if isinstance(value, str):
            v1, colon, v2 = value.partition(":")
            try:
                v1 = int(v1)
                v2 = int(v2) if colon else None
            except ValueError:
                self.fail("Value must be of the form `N[:M]`", param, ctx)
            return (v1, v2)
        else:
            return value

    def get_metavar(self, param):
        return "N[:M]"


@click.command()
# @dandiset_path_option(
#     help="Top directory (local) of the dandiset.  Files will be uploaded with "
#     "paths relative to that directory. If not specified, current or a parent "
#     "directory containing dandiset.yaml file will be assumed "
# )
@click.option(
    "-e",
    "--existing",
    type=click.Choice(["error", "skip", "force", "overwrite", "refresh"]),
    help="What to do if a file found existing on the server. 'skip' would skip"
    "the file, 'force' - force reupload, 'overwrite' - force upload if "
    "either size or modification time differs; 'refresh' - upload only if "
    "local modification time is ahead of the remote.",
    default="refresh",
    show_default=True,
)
@click.option(
    "-J",
    "--jobs",
    type=IntColonInt(),
    help="Number of files to upload in parallel and, optionally, number of upload threads per file",
)
@click.option(
    "--sync", is_flag=True, help="Delete assets on the server that do not exist locally"
)
@click.option(
    "--validation",
    help="Data must pass validation before the upload.  Use of this option is highly discouraged.",
    type=click.Choice(["require", "skip", "ignore"]),
    default="require",
    show_default=True,
)
@click.argument("paths", nargs=-1)  # , type=click.Path(exists=True, dir_okay=False))
# &
# Development options:  Set DANDI_DEVEL for them to become available
#
# TODO: should always go to dandi for now
@instance_option()
@devel_option(
    "--allow-any-path",
    help="For development: allow DANDI 'unsupported' file types/paths",
    default=False,
    is_flag=True,
)
@devel_option(
    "--upload-dandiset-metadata",
    help="For development: do upload dandiset metadata",
    default=False,
    is_flag=True,
)
@devel_debug_option()
@map_to_click_exceptions
def upload(
    paths,
    jobs,
    sync,
    existing="refresh",
    validation="require",
    dandiset_path=None,
    # Development options should come as kwargs
    dandi_instance="dandi",
    allow_any_path=False,
    upload_dandiset_metadata=False,
    devel_debug=False,
):
    """Upload dandiset (files) to DANDI archive.

    Target dandiset to upload to must already be registered in the archive and
    locally "dandiset.yaml" should exist in `--dandiset-path`.  If you have not
    yet created a dandiset in the archive, use 'dandi register' command first.

    Local dandiset should pass validation.  For that it should be first organized
    using 'dandiset organize' command.

    By default all files in the dandiset (not following directories starting with a period)
    will be considered for the upload.  You can point to specific files you would like to
    validate and have uploaded.
    """
    from ..upload import upload

    if jobs is None:
        jobs = None
        jobs_per_file = None
    else:
        jobs, jobs_per_file = jobs

    upload(
        paths,
        existing=existing,
        validation=validation,
        dandiset_path=dandiset_path,
        dandi_instance=dandi_instance,
        allow_any_path=allow_any_path,
        upload_dandiset_metadata=upload_dandiset_metadata,
        devel_debug=devel_debug,
        jobs=jobs,
        jobs_per_file=jobs_per_file,
        sync=sync,
    )



import re

import pynwb

from ..pynwb_utils import _sanitize_nwb_version


def test_pynwb_io(simple1_nwb):
    # To verify that our dependencies spec is sufficient to avoid
    # stepping into known pynwb/hdmf issues
    with pynwb.NWBHDF5IO(str(simple1_nwb), "r", load_namespaces=True) as reader:
        nwbfile = reader.read()
    assert repr(nwbfile)
    assert str(nwbfile)


def test_sanitize_nwb_version():
    def _nocall(*args):
        raise AssertionError(f"Should have not been called. Was called with {args}")

    def assert_regex(regex):
        def search(v):
            assert re.search(regex, v)

        return search

    assert _sanitize_nwb_version("1.0.0", log=_nocall) == "1.0.0"
    assert _sanitize_nwb_version("NWB-1.0.0", log=_nocall) == "1.0.0"
    assert _sanitize_nwb_version("NWB-2.0.0", log=_nocall) == "2.0.0"
    assert (
        _sanitize_nwb_version(
            "NWB-2.1.0",
            log=assert_regex("^nwb_version 'NWB-2.1.0' starts with NWB- prefix,"),
        )
        == "2.1.0"
    )
    assert (
        _sanitize_nwb_version(
            "NWB-2.1.0",
            filename="/bu",
            log=assert_regex(
                "^File /bu: nwb_version 'NWB-2.1.0' starts with NWB- prefix,"
            ),
        )
        == "2.1.0"
    )



import json
import os
import os.path as op
from pathlib import Path
from shutil import rmtree

import pytest

from ..download import download
from ..utils import find_files


# both urls point to 000027 (lean test dataset), and both draft and "released"
# version have only a single file ATM
@pytest.mark.parametrize(
    "url",
    [  # Should go through API
        pytest.param(
            "https://dandiarchive.org/dandiset/000027/0.200721.2222",
            marks=pytest.mark.xfail(reason="publish.dandiarchive.org is gone"),
        ),
        # Drafts do not go through API ATM, but that should not be visible to user
        "https://dandiarchive.org/dandiset/000027/draft",
    ],
)
def test_download_000027(url, tmpdir):
    ret = download(url, tmpdir)
    assert not ret  # we return nothing ATM, might want to "generate"
    dsdir = tmpdir / "000027"
    downloads = (x.relto(dsdir) for x in dsdir.visit())
    assert sorted(downloads) == [
        "dandiset.yaml",
        "sub-RAT123",
        op.join("sub-RAT123", "sub-RAT123.nwb"),
    ]
    # and checksum should be correct as well
    from ..support.digests import Digester

    assert (
        Digester(["md5"])(dsdir / "sub-RAT123" / "sub-RAT123.nwb")["md5"]
        == "33318fd510094e4304868b4a481d4a5a"
    )
    # redownload - since already exist there should be an exception
    with pytest.raises(FileExistsError):
        download(url, tmpdir)

    # TODO: somehow get that status report about what was downloaded and what not
    download(url, tmpdir, existing="skip")  # TODO: check that skipped
    download(url, tmpdir, existing="overwrite")  # TODO: check that redownloaded
    download(url, tmpdir, existing="refresh")  # TODO: check that skipped (the same)


@pytest.mark.parametrize(
    "url",
    [  # Should go through API
        pytest.param(
            "https://dandiarchive.org/dandiset/000027/0.200721.2222",
            marks=pytest.mark.xfail(reason="publish.dandiarchive.org is gone"),
        ),
        # Drafts do not go through API ATM, but that should not be visible to user
        "https://dandiarchive.org/dandiset/000027/draft",
    ],
)
def test_download_000027_metadata_only(url, tmpdir):
    ret = download(url, tmpdir, get_assets=False)
    assert not ret  # we return nothing ATM, might want to "generate"
    dsdir = tmpdir / "000027"
    downloads = (x.relto(dsdir) for x in dsdir.visit())
    assert sorted(downloads) == ["dandiset.yaml"]


@pytest.mark.parametrize(
    "url",
    [  # Should go through API
        pytest.param(
            "https://dandiarchive.org/dandiset/000027/0.200721.2222",
            marks=pytest.mark.xfail(reason="publish.dandiarchive.org is gone"),
        ),
        # Drafts do not go through API ATM, but that should not be visible to user
        "https://dandiarchive.org/dandiset/000027/draft",
    ],
)
def test_download_000027_assets_only(url, tmpdir):
    ret = download(url, tmpdir, get_metadata=False)
    assert not ret  # we return nothing ATM, might want to "generate"
    dsdir = tmpdir / "000027"
    downloads = (x.relto(dsdir) for x in dsdir.visit())
    assert sorted(downloads) == ["sub-RAT123", op.join("sub-RAT123", "sub-RAT123.nwb")]


@pytest.mark.parametrize("resizer", [lambda sz: 0, lambda sz: sz // 2, lambda sz: sz])
@pytest.mark.parametrize(
    "version",
    [
        pytest.param(
            "0.200721.2222",
            marks=pytest.mark.xfail(reason="publish.dandiarchive.org is gone"),
        ),
        "draft",
    ],
)
def test_download_000027_resume(tmp_path, resizer, version):
    from ..support.digests import Digester

    url = f"https://dandiarchive.org/dandiset/000027/{version}"
    digester = Digester()
    download(url, tmp_path, get_metadata=False)
    dsdir = tmp_path / "000027"
    nwb = dsdir / "sub-RAT123" / "sub-RAT123.nwb"
    digests = digester(str(nwb))
    dldir = nwb.with_name(nwb.name + ".dandidownload")
    dldir.mkdir()
    dlfile = dldir / "file"
    nwb.rename(dlfile)
    size = dlfile.stat().st_size
    os.truncate(dlfile, resizer(size))
    with (dldir / "checksum").open("w") as fp:
        json.dump(digests, fp)
    download(url, tmp_path, get_metadata=False)
    contents = [
        op.relpath(op.join(dirpath, entry), dsdir)
        for (dirpath, dirnames, filenames) in os.walk(dsdir)
        for entry in dirnames + filenames
    ]
    assert sorted(contents) == ["sub-RAT123", op.join("sub-RAT123", "sub-RAT123.nwb")]
    assert nwb.stat().st_size == size
    assert digester(str(nwb)) == digests


def test_download_newest_version(local_dandi_api, text_dandiset, tmp_path):
    dandiset_id = text_dandiset["dandiset_id"]
    download(f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}", tmp_path)
    assert (tmp_path / dandiset_id / "file.txt").read_text() == "This is test text.\n"
    text_dandiset["client"].publish_version(dandiset_id, "draft")
    (text_dandiset["dspath"] / "file.txt").write_text("This is different text.\n")
    text_dandiset["reupload"]()
    rmtree(tmp_path / dandiset_id)
    download(f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}", tmp_path)
    assert (tmp_path / dandiset_id / "file.txt").read_text() == "This is test text.\n"


def test_download_folder(local_dandi_api, text_dandiset, tmp_path):
    dandiset_id = text_dandiset["dandiset_id"]
    download(
        f"dandi://{local_dandi_api['instance_id']}/{dandiset_id}/subdir2/", tmp_path
    )
    assert sorted(map(Path, find_files(r".*", paths=[tmp_path], dirs=True))) == [
        tmp_path / "subdir2",
        tmp_path / "subdir2" / "banana.txt",
        tmp_path / "subdir2" / "coconut.txt",
    ]
    assert (tmp_path / "subdir2" / "banana.txt").read_text() == "Banana\n"
    assert (tmp_path / "subdir2" / "coconut.txt").read_text() == "Coconut\n"


def test_download_item(local_dandi_api, text_dandiset, tmp_path):
    dandiset_id = text_dandiset["dandiset_id"]
    download(
        f"dandi://{local_dandi_api['instance_id']}/{dandiset_id}/subdir2/coconut.txt",
        tmp_path,
    )
    assert list(map(Path, find_files(r".*", paths=[tmp_path], dirs=True))) == [
        tmp_path / "coconut.txt"
    ]
    assert (tmp_path / "coconut.txt").read_text() == "Coconut\n"


def test_download_asset_id(local_dandi_api, text_dandiset, tmp_path):
    dandiset_id = text_dandiset["dandiset_id"]
    asset = local_dandi_api["client"].get_asset_bypath(
        dandiset_id, "draft", "subdir2/coconut.txt"
    )
    assert asset is not None
    download(
        f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}/versions"
        f"/draft/assets/{asset['asset_id']}/download/",
        tmp_path,
    )
    assert list(map(Path, find_files(r".*", paths=[tmp_path], dirs=True))) == [
        tmp_path / "coconut.txt"
    ]
    assert (tmp_path / "coconut.txt").read_text() == "Coconut\n"


@pytest.mark.parametrize("confirm", [True, False])
def test_download_sync(confirm, local_dandi_api, mocker, text_dandiset, tmp_path):
    text_dandiset["client"].delete_asset_bypath(
        text_dandiset["dandiset_id"], "draft", "file.txt"
    )
    dspath = tmp_path / text_dandiset["dandiset_id"]
    os.rename(text_dandiset["dspath"], dspath)
    confirm_mock = mocker.patch(
        "dandi.download.abbrev_prompt", return_value="yes" if confirm else "no"
    )
    download(
        f"dandi://{local_dandi_api['instance_id']}/{text_dandiset['dandiset_id']}",
        tmp_path,
        existing="overwrite",
        sync=True,
    )
    confirm_mock.assert_called_with("Delete 1 local asset?", "yes", "no", "list")
    if confirm:
        assert not (dspath / "file.txt").exists()
    else:
        assert (dspath / "file.txt").exists()


def test_download_sync_folder(local_dandi_api, mocker, text_dandiset):
    text_dandiset["client"].delete_asset_bypath(
        text_dandiset["dandiset_id"], "draft", "file.txt"
    )
    text_dandiset["client"].delete_asset_bypath(
        text_dandiset["dandiset_id"], "draft", "subdir2/banana.txt"
    )
    confirm_mock = mocker.patch("dandi.download.abbrev_prompt", return_value="yes")
    download(
        f"dandi://{local_dandi_api['instance_id']}/{text_dandiset['dandiset_id']}/subdir2/",
        text_dandiset["dspath"],
        existing="overwrite",
        sync=True,
    )
    confirm_mock.assert_called_with("Delete 1 local asset?", "yes", "no", "list")
    assert (text_dandiset["dspath"] / "file.txt").exists()
    assert not (text_dandiset["dspath"] / "subdir2" / "banana.txt").exists()


def test_download_sync_list(capsys, local_dandi_api, mocker, text_dandiset, tmp_path):
    text_dandiset["client"].delete_asset_bypath(
        text_dandiset["dandiset_id"], "draft", "file.txt"
    )
    dspath = tmp_path / text_dandiset["dandiset_id"]
    os.rename(text_dandiset["dspath"], dspath)
    input_mock = mocker.patch("dandi.utils.input", side_effect=["list", "yes"])
    download(
        f"dandi://{local_dandi_api['instance_id']}/{text_dandiset['dandiset_id']}",
        tmp_path,
        existing="overwrite",
        sync=True,
    )
    assert not (dspath / "file.txt").exists()
    assert input_mock.call_args_list == [
        mocker.call("Delete 1 local asset? ([y]es/[n]o/[l]ist): "),
        mocker.call("Delete 1 local asset? ([y]es/[n]o/[l]ist): "),
    ]
    assert capsys.readouterr().out.splitlines()[-1] == str(dspath / "file.txt")



import hashlib
import json
import os
import os.path as op
from pathlib import Path
import random
from shutil import rmtree
import sys
import time

import humanize
import requests

from .consts import dandiset_metadata_file
from .dandiarchive import DandisetURL, MultiAssetURL, SingleAssetURL, parse_dandi_url
from .dandiset import Dandiset
from . import get_logger
from .support.pyout import naturalsize
from .utils import (
    abbrev_prompt,
    ensure_datetime,
    find_files,
    flattened,
    is_same_time,
    on_windows,
    pluralize,
)

lgr = get_logger()


def download(
    urls,
    output_dir,
    *,
    format="pyout",
    existing="error",
    jobs=1,
    get_metadata=True,
    get_assets=True,
    sync=False,
):
    # TODO: unduplicate with upload. For now stole from that one
    # We will again use pyout to provide a neat table summarizing our progress
    # with upload etc
    from .support import pyout as pyouts

    urls = flattened([urls])
    if len(urls) > 1:
        raise NotImplementedError("multiple URLs not supported")
    if not urls:
        # if no paths provided etc, we will download dandiset path
        # we are at, BUT since we are not git -- we do not even know
        # on which instance it exists!  Thus ATM we would do nothing but crash
        raise NotImplementedError("No URLs were provided.  Cannot download anything")

    parsed_url = parse_dandi_url(urls[0])

    # TODO: if we are ALREADY in a dandiset - we can validate that it is the
    # same dandiset and use that dandiset path as the one to download under
    if isinstance(parsed_url, DandisetURL):
        output_path = op.join(output_dir, parsed_url.dandiset_id)
    else:
        output_path = output_dir

    # dandi.cli.formatters are used in cmd_ls to provide switchable
    pyout_style = pyouts.get_style(hide_if_missing=False)

    rec_fields = ("path", "size", "done", "done%", "checksum", "status", "message")
    out = pyouts.LogSafeTabular(style=pyout_style, columns=rec_fields, max_workers=jobs)

    out_helper = PYOUTHelper()
    pyout_style["done"] = pyout_style["size"].copy()
    pyout_style["size"]["aggregate"] = out_helper.agg_size
    pyout_style["done"]["aggregate"] = out_helper.agg_done

    # I thought I was making a beautiful flower but ended up with cacti
    # which never blooms... All because assets are looped through inside download_generator
    # TODO: redo
    kw = dict(assets_it=out_helper.it)
    if jobs > 1 and format == "pyout":
        # It could handle delegated to generator downloads
        kw["yield_generator_for_fields"] = rec_fields[1:]  # all but path

    gen_ = download_generator(
        parsed_url,
        output_path,
        existing=existing,
        get_metadata=get_metadata,
        get_assets=get_assets,
        **kw,
    )

    # TODOs:
    #  - redo frontends similarly to how command_ls did it
    #  - have a single loop with analysis of `rec` to either any file
    #    has failed to download.  If any was: exception should probably be
    #    raised.  API discussion for Python side of API:
    #
    if format == "debug":
        for rec in gen_:
            print(rec)
            sys.stdout.flush()
    elif format == "pyout":
        with out:
            for rec in gen_:
                out(rec)
    else:
        raise ValueError(format)

    if sync and not isinstance(parsed_url, SingleAssetURL):
        client = parsed_url.get_client()
        with client.session():
            asset_paths = {asset["path"] for asset in parsed_url.get_assets(client)}
        if isinstance(parsed_url, DandisetURL):
            prefix = os.curdir
            download_dir = output_path
        elif isinstance(parsed_url, MultiAssetURL):
            folder_path = op.normpath(parsed_url.path)
            prefix = folder_path
            download_dir = op.join(output_path, op.basename(folder_path))
        else:
            raise NotImplementedError(
                f"Unexpected URL type {type(parsed_url).__name__}"
            )
        to_delete = []
        for p in find_files(".*", download_dir, exclude_datalad=True):
            if p == op.join(output_path, dandiset_metadata_file):
                continue
            a_path = op.normpath(op.join(prefix, op.relpath(p, download_dir)))
            if on_windows:
                a_path = a_path.replace("\\", "/")
            if a_path not in asset_paths:
                to_delete.append(p)
        if to_delete:
            while True:
                opt = abbrev_prompt(
                    f"Delete {pluralize(len(to_delete), 'local asset')}?",
                    "yes",
                    "no",
                    "list",
                )
                if opt == "list":
                    for p in to_delete:
                        print(p)
                elif opt == "yes":
                    for p in to_delete:
                        os.unlink(p)
                    break
                else:
                    break


def download_generator(
    parsed_url,
    output_path,
    *,
    assets_it=None,
    yield_generator_for_fields=None,
    existing="error",
    get_metadata=True,
    get_assets=True,
):
    """A generator for downloads of files, folders, or entire dandiset from DANDI
    (as identified by URL)

    This function is a generator which would yield records on ongoing activities.
    Activities include traversal of the remote resource (DANDI archive), download of
    individual assets while yielding records (TODO: schema) while validating their
    checksums "on the fly", etc.

    Parameters
    ----------
    assets_it: IteratorWithAggregation
      which will be set .gen to assets.  Purpose is to make it possible to get
      summary statistics while already downloading.  TODO: reimplement properly!

    """

    with parsed_url.navigate() as (client, dandiset, assets):
        if assets_it:
            assets_it.gen = assets
            assets = assets_it

        if isinstance(parsed_url, DandisetURL) and get_metadata:
            for resp in _populate_dandiset_yaml(output_path, dandiset, existing):
                yield dict(path=dandiset_metadata_file, **resp)

        # TODO: do analysis of assets for early detection of needed renames etc
        # to avoid any need for late treatment of existing and also for
        # more efficient download if files are just renamed etc

        if not get_assets:
            return

        for asset in assets:
            down_args = (
                dandiset["dandiset"]["identifier"],
                dandiset["version"],
                asset["asset_id"],
            )
            metadata = asset["metadata"]
            d = metadata.get("digest", {})
            if "dandi:dandi-etag" in d:
                digests = {"dandi-etag": d["dandi:dandi-etag"]}
            else:
                raise RuntimeError(
                    f"dandi-etag not available for asset. Known digests: {d}"
                )

            path = asset["path"].lstrip("/")  # make into relative path
            path = op.normpath(path)
            if not isinstance(parsed_url, DandisetURL):
                if isinstance(parsed_url, MultiAssetURL):
                    folder_path = op.normpath(parsed_url.path)
                    path = op.join(
                        op.basename(folder_path), op.relpath(path, folder_path)
                    )
                elif isinstance(parsed_url, SingleAssetURL):
                    path = op.basename(path)
                else:
                    raise NotImplementedError(
                        f"Unexpected URL type {type(parsed_url).__name__}"
                    )
            download_path = op.join(output_path, path)

            downloader = client.get_download_file_iter(*down_args)

            try:
                mtime = asset["metadata"]["blobDateModified"]
            except KeyError:
                mtime = None
            if mtime is None:
                lgr.warning(
                    "Asset %s is missing blobDateModified metadata field",
                    asset["path"],
                )
                mtime = asset.get("modified")

            _download_generator = _download_file(
                downloader,
                download_path,
                # size and modified generally should be there but better to redownload
                # than to crash
                size=asset.get("size"),
                mtime=mtime,
                existing=existing,
                digests=digests,
            )

            if yield_generator_for_fields:
                yield {"path": path, yield_generator_for_fields: _download_generator}
            else:
                for resp in _download_generator:
                    yield dict(resp, path=path)


class ItemsSummary:
    """A helper "structure" to accumulate information about assets to be downloaded

    To be used as a callback to IteratorWithAggregation
    """

    def __init__(self):
        self.files = 0
        # TODO: get rid of needing it
        self.t0 = None  # when first record is seen
        self.size = 0
        self.has_unknown_sizes = False

    def as_dict(self):
        return {a: getattr(self, a) for a in ("files", "size", "has_unknown_sizes")}

    def __call__(self, rec, prior=None):
        assert prior in (None, self)
        if not self.files:
            self.t0 = time.time()
        self.files += 1
        size = rec.get("size")
        if size is not None:
            self.size += size
        elif rec.get("path", "") == "dandiset.yaml":
            # again -- so special. TODO: make it a proper file
            pass
        else:
            self.has_unknown_sizes = True
        return self


class PYOUTHelper:
    """Helper for PYOUT styling

    Provides aggregation callbacks for PyOUT and also an iterator to be wrapped around
    iterating over assets, so it would get "totals" as soon as they are available.
    """

    def __init__(self):
        # Establish "fancy" download while still possibly traversing the dandiset
        # functionality.
        from .support.iterators import IteratorWithAggregation

        self.items_summary = ItemsSummary()
        self.it = IteratorWithAggregation(
            # unfortunately Yarik missed the point that we need to wrap
            # "assets" generator within downloader_generator
            # so we do not have assets here!  Ad-hoc solution for now is to
            # pass this beast so it could get .gen set within downloader_generator
            None,  # download_generator(urls, output_dir, existing=existing),
            self.items_summary,
        )

    def agg_files(self, *ignored):
        ret = str(self.items_summary.files)
        if not self.it.finished:
            ret += "+"
        return ret

    def agg_size(self, sizes):
        """Formatter for "size" column where it would show

        how much is "active" (or done)
        +how much yet to be "shown".
        """
        active = sum(sizes)
        if (active, self.items_summary.size) == (0, 0):
            return ""
        v = [naturalsize(active)]
        if not self.it.finished or (
            active != self.items_summary.size or self.items_summary.has_unknown_sizes
        ):
            extra = self.items_summary.size - active
            if extra < 0:
                lgr.debug("Extra size %d < 0 -- must not happen", extra)
            else:
                extra_str = "+%s" % naturalsize(extra)
                if not self.it.finished:
                    extra_str = ">" + extra_str
                if self.items_summary.has_unknown_sizes:
                    extra_str += "+?"
                v.append(extra_str)
        return v

    def agg_done(self, done_sizes):
        """Formatter for "DONE" column"""
        done = sum(done_sizes)
        if self.it.finished and done == 0 and self.items_summary.size == 0:
            # even with 0s everywhere consider it 100%
            r = 1.0
        elif self.items_summary.size:
            r = done / self.items_summary.size
        else:
            r = 0
        pref = ""
        if not self.it.finished:
            pref += "<"
        if self.items_summary.has_unknown_sizes:
            pref += "?"
        v = [naturalsize(done), "%s%.2f%%" % (pref, 100 * r)]
        if (
            done
            and self.items_summary.t0 is not None
            and r
            and self.items_summary.size != 0
        ):
            dt = time.time() - self.items_summary.t0
            more_time = dt / r if r != 1 else 0
            more_time_str = humanize.naturaldelta(more_time)
            if not self.it.finished:
                more_time_str += "<"
            if self.items_summary.has_unknown_sizes:
                more_time_str += "+?"
            if more_time:
                v.append("ETA: %s" % more_time_str)
        return v


def _skip_file(msg):
    return {"status": "skipped", "message": str(msg)}


def _populate_dandiset_yaml(dandiset_path, dandiset_data, existing):
    metadata = dandiset_data.get("metadata")
    if not metadata:
        lgr.warning(
            "Got completely empty metadata record for dandiset, not producing dandiset.yaml"
        )
        return
    dandiset_yaml = op.join(dandiset_path, dandiset_metadata_file)
    yield {"message": "updating"}
    lgr.debug("Updating %s from obtained dandiset metadata", dandiset_metadata_file)
    mtime = ensure_datetime(dandiset_data["modified"])
    if op.lexists(dandiset_yaml):
        if existing == "error":
            raise FileExistsError(dandiset_yaml)
        elif existing == "skip" or (
            existing == "refresh"
            and os.lstat(dandiset_yaml).st_mtime >= mtime.timestamp()
        ):
            yield _skip_file("already exists")
            return
    dandiset = Dandiset(dandiset_path, allow_empty=True)
    dandiset.path_obj.mkdir(exist_ok=True)  # exist_ok in case of parallel race
    old_metadata = dandiset.metadata
    dandiset.update_metadata(metadata)
    os.utime(dandiset_yaml, (time.time(), mtime.timestamp()))
    yield {
        "status": "done",
        "message": "updated" if metadata != old_metadata else "same",
    }


def _download_file(
    downloader, path, size=None, mtime=None, existing="error", digests=None
):
    """Common logic for downloading a single file

    Generator downloader:

    TODO: describe expected records it should yield
    - progress
    - error
    - completion

    Parameters
    ----------
    downloader: callable returning a generator
      A backend-specific fixture for downloading some file into path. It should
      be a generator yielding downloaded blocks.
    size: int, optional
      Target size if known
    digests: dict, optional
      possible checksums or other digests provided for the file. Only one
      will be used to verify download
    """
    if op.lexists(path):
        block = f"File {path!r} already exists"
        if existing == "error":
            raise FileExistsError(block)
        elif existing == "skip":
            yield _skip_file("already exists")
            return
        elif existing == "overwrite":
            pass
        elif existing == "refresh":
            if mtime is None:
                lgr.warning(
                    f"{path!r} - no mtime or ctime in the record, redownloading"
                )
            else:
                stat = os.stat(op.realpath(path))
                same = []
                if is_same_time(stat.st_mtime, mtime):
                    same.append("mtime")
                if size is not None and stat.st_size == size:
                    same.append("size")
                # TODO: use digests if available? or if e.g. size is identical
                # but mtime is different
                if same == ["mtime", "size"]:
                    # TODO: add recording and handling of .nwb object_id
                    yield _skip_file("same time and size")
                    return
                lgr.debug(f"{path!r} - same attributes: {same}.  Redownloading")

    if size is not None:
        yield {"size": size}

    destdir = op.dirname(path)
    os.makedirs(destdir, exist_ok=True)

    yield {"status": "downloading"}

    algo, digester, digest, downloaded_digest = None, None, None, None
    if digests:
        # choose first available for now.
        # TODO: reuse that sorting based on speed
        for algo, digest in digests.items():
            if algo == "dandi-etag":
                from .core.digests.dandietag import ETagHashlike

                digester = lambda: ETagHashlike(size)  # noqa: E731
            else:
                digester = getattr(hashlib, algo, None)
            if digester:
                break
        if not digester:
            lgr.warning("Found no digests in hashlib for any of %s", str(digests))

    # TODO: how do we discover the total size????
    # TODO: do not do it in-place, but rather into some "hidden" file
    for attempt in range(3):
        try:
            if digester:
                downloaded_digest = digester()  # start empty
            warned = False
            # I wonder if we could make writing async with downloader
            with DownloadDirectory(path, digests) as dldir:
                downloaded = dldir.offset
                if size is not None and downloaded == size:
                    # Exit early when downloaded == size, as making a Range
                    # request in such a case results in a 416 error from S3.
                    # Problems will result if `size` is None but we've already
                    # downloaded everything.
                    break
                for block in downloader(start_at=dldir.offset):
                    if digester:
                        downloaded_digest.update(block)
                    downloaded += len(block)
                    # TODO: yield progress etc
                    msg = {"done": downloaded}
                    if size:
                        if downloaded > size and not warned:
                            warned = True
                            # Yield ERROR?
                            lgr.warning(
                                "Downloaded %d bytes although size was told to be just %d",
                                downloaded,
                                size,
                            )
                        msg["done%"] = 100 * downloaded / size if size else "100"
                        # TODO: ETA etc
                    yield msg
                    dldir.append(block)
            break
        except requests.exceptions.HTTPError as exc:
            # TODO: actually we should probably retry only on selected codes, and also
            # respect Retry-After
            if attempt >= 2 or exc.response.status_code not in (
                400,  # Bad Request, but happened with gider:
                # https://github.com/dandi/dandi-cli/issues/87
                503,  # Service Unavailable
            ):
                lgr.debug("Download failed: %s", exc)
                yield {"status": "error", "message": str(exc)}
                return
            # if is_access_denied(exc) or attempt >= 2:
            #     raise
            # sleep a little and retry
            lgr.debug(
                "Failed to download on attempt#%d: %s, will sleep a bit and retry",
                attempt,
                exc,
            )
            time.sleep(random.random() * 5)

    if downloaded_digest:
        downloaded_digest = downloaded_digest.hexdigest()  # we care only about hex
        if digest != downloaded_digest:
            msg = f"{algo}: downloaded {downloaded_digest} != {digest}"
            yield {"checksum": "differs", "status": "error", "message": msg}
            lgr.debug("%s is different: %s.", path, msg)
            return
        else:
            yield {"checksum": "ok"}
            lgr.debug("Verified that %s has correct %s %s", path, algo, digest)
    else:
        # shouldn't happen with more recent metadata etc
        yield {
            "checksum": "-",
            # "message": "no digests were provided"
        }

    # TODO: dissolve attrs and pass specific mtime?
    if mtime:
        yield {"status": "setting mtime"}
        os.utime(path, (time.time(), ensure_datetime(mtime).timestamp()))

    yield {"status": "done"}


class DownloadDirectory:
    def __init__(self, filepath, digests):
        #: The path to which to save the file after downloading
        self.filepath = Path(filepath)
        #: Expected hashes of the downloaded data, as a mapping from algorithm
        #: names to digests
        self.digests = digests
        #: The working directory in which downloaded data will be temporarily
        #: stored
        self.dirpath = self.filepath.with_name(self.filepath.name + ".dandidownload")
        #: The file in `dirpath` to which data will be written as it is
        #: received
        self.writefile = self.dirpath / "file"
        #: A `fasteners.InterProcessLock` on `dirpath`
        self.lock = None
        #: An open filehandle to `writefile`
        self.fp = None
        #: How much of the data has been downloaded so far
        self.offset = None

    def __enter__(self):
        from fasteners import InterProcessLock

        self.dirpath.mkdir(parents=True, exist_ok=True)
        self.lock = InterProcessLock(str(self.dirpath / "lock"))
        if not self.lock.acquire(blocking=False):
            raise RuntimeError("Could not acquire download lock for {self.filepath}")
        chkpath = self.dirpath / "checksum"
        try:
            with chkpath.open() as fp:
                digests = json.load(fp)
        except (FileNotFoundError, ValueError):
            digests = {}
        matching_algs = self.digests.keys() & digests.keys()
        if matching_algs and all(
            self.digests[alg] == digests[alg] for alg in matching_algs
        ):
            # Pick up where we left off, writing to the end of the file
            lgr.debug(
                "Download directory exists and has matching checksum; resuming download"
            )
            self.fp = self.writefile.open("ab")
        else:
            # Delete the file (if it even exists) and start anew
            if not chkpath.exists():
                lgr.debug("Starting new download in new download directory")
            else:
                lgr.debug(
                    "Download directory found, but digests do not match; starting new download"
                )
            try:
                self.writefile.unlink()
            except FileNotFoundError:
                pass
            self.fp = self.writefile.open("wb")
        with chkpath.open("w") as fp:
            json.dump(self.digests, fp)
        self.offset = self.fp.tell()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.fp.close()
        try:
            if exc_type is None:
                self.writefile.replace(self.filepath)
        finally:
            self.lock.release()
            if exc_type is None:
                rmtree(self.dirpath, ignore_errors=True)
            self.lock = None
            self.fp = None
            self.offset = None
        return False

    def append(self, blob):
        self.fp.write(blob)
